import os
import re

import pkg_resources


def write_vmd(
    colors,  # array of rgb values
    trajectories,  # array of filenames
    states,
    opt_in={},  # dictionary of options
):

    """Writes vmd script to visualize trajectories

    Params:
        colors, np.ndarray (nstates,3), array of rgb values for each state
        trajectories, list (ntrajectories), list of xyzs filenames
        states, dictionary (ntrajectories), trajectory index to state label (0 based)
        opt_in, dictionary (options to override defaults)

    """

    header_path = pkg_resources.resource_filename(__name__, "ref/header.vmd")
    footer_path = pkg_resources.resource_filename(__name__, "ref/footer.vmd")
    molecule_path = pkg_resources.resource_filename(__name__, "ref/molecule.vmd")

    # Set User Options
    options = {
        "fn_header": header_path,
        "fn_footer": footer_path,
        "fn_mol": molecule_path,
        "fn_vmd": "vis_fms.vmd",
        "render": True,
        "exit_vmd": True,
        "opacities": True,
    }
    for key, val in list(opt_in.items()):
        options[key] = val

    fn_header = options["fn_header"]
    fn_footer = options["fn_footer"]
    fn_mol = options["fn_mol"]
    fn_vmd = options["fn_vmd"]
    options["render"]
    options["opacities"]
    exit_vmd = options["exit_vmd"]  # whether to render and exit vmd

    # Lines from vmd Templates
    header_lines = open(fn_header).readlines()
    footer_lines = open(fn_footer).readlines()
    mol_lines = open(fn_mol).readlines()

    # Number of trajectories
    ntrajs = len(trajectories)

    # Write the header
    # Write new material for each opacity
    # default to 1.0/ntrajs?
    fh = open(fn_vmd, "w")
    for lh in header_lines:
        mobj = re.match(r"^  # Add materials!\s*", lh)
        mobj1 = re.match(
            r"^  set mlist { Opaque Transparent BrushedMetal Diffuse Ghost Glass1 Glass2 Glass3 Glossy HardPlastic MetallicPastel Steel Translucent Edgy EdgyShiny EdgyGlass Goodsell AOShiny AOChalky AOEdgy BlownGlass GlassBubble RTChrome }",
            lh,
        )
        if mobj:
            fh.write(lh)
            for n in range(ntrajs):
                fh.write("  material change ambient Transp%d 0.000000\n" % (n + 1))
                fh.write("  material change diffuse Transp%d 0.920000\n" % (n + 1))
                fh.write("  material change specular Transp%d 0.290000\n" % (n + 1))
                fh.write("  material change shininess Transp%d 0.470000\n" % (n + 1))
                fh.write("  material change mirror Transp%d 0.000000\n" % (n + 1))
                fh.write("  material change opacity Transp%d 0.500000\n" % (n + 1))
                fh.write("  material change outline Transp%d 1.510000\n" % (n + 1))
        if mobj1:
            fh.write(
                "  set mlist { Opaque Transparent BrushedMetal Diffuse Ghost Glass1 Glass2 Glass3 Glossy HardPlastic MetallicPastel Steel Translucent Edgy EdgyShiny EdgyGlass Goodsell AOShiny AOChalky AOEdgy BlownGlass GlassBubble RTChrome "
            )
            for n in range(ntrajs):
                fh.write("Transp%d " % (n + 1))
            fh.write("}\n")
        else:
            fh.write(lh)

    # Looping through trajectories
    1.0 / ntrajs
    for n in range(ntrajs):
        # Writing vmd segment for each trajecotry
        for ml in mol_lines:
            # Read in filename
            mobj1 = re.sub(r"filename", trajectories[n], ml)
            # Assign color ID number
            mobj2 = re.sub(
                r"mol color ColorID \d+", r"mol color ColorID %d" % (32 - n), mobj1
            )
            # Assign material to representation
            mobj3 = re.sub(
                r"mol material Transp\d+", r"mol material Transp%d" % (n + 1), mobj2
            )
            # Reassign RGB value to color ID
            mobj4 = re.sub(
                r"color change rgb \d+ \d\.\d+ \d\.\d+ \d\.\d+",
                r"color change rgb %d %6.5f %6.5f %6.5f"
                % (
                    32 - n,
                    colors[states[n]][0],
                    colors[states[n]][1],
                    colors[states[n]][2],
                ),
                mobj3,
            )
            fh.write("%s" % (mobj4))
        fh.write("# Finished traj %d\n" % (n + 1))

    # Writing footer for vmd file
    for lf in footer_lines:
        fh.write(lf)
    if exit_vmd:
        fh.write("source render.tcl\n")
        fh.write("exit")


def write_render(
    opt_in={},
):

    """Renders each frame of trajectory in vmd

    Params:
        opt_in, dictionary (options to override defaults)

    """

    options = {
        # visualization options
        "opacities": [],
        "end_frame": 1000,
        "increment_frame": 10,
        "sleep": 1,
        "ntrajs": 0,
        "display_size": None,
        "translation": None,
        "axis_on": False,
        "scale": 0.25,
    }
    for key, val in list(opt_in.items()):
        options[key] = val
    opacities = options["opacities"]
    end_frame = options["end_frame"]
    increment_frame = options["increment_frame"]
    ntrajs = options["ntrajs"]
    sleep = options["sleep"]
    display_size = options["display_size"]  # (1000, 600)
    translation = options["translation"]  # [-0.5, 0.0, 0.0]
    axis_on = options["axis_on"]
    scale = options["scale"]

    # TODO: Create options for:
    # Length of trajectory
    # Interval of frames (steps)
    # Translations and rotations
    # image scale and size
    # Render sleep settings

    # Number of trajectories
    fh = open("render.tcl", "w")

    # Writing Opacities based on amplitude for all trajectories
    # Setting the amplitude to zero for the time before Spawning
    # For groundstate trajectories that dropped, setting the amplitude
    # to it's last amplitude until the end of the trajectory
    for n in range(ntrajs):
        fh.write("set opa%d [list   " % (n + 1))
        for val in opacities[n]:
            fh.write("%4.3f   " % (val))
        fh.write("]\n")

    # Writing remainder of display settings
    if not axis_on:
        fh.write("axes location off\n")
    # TODO: include custom image size settings!
    fh.write("scale to %1.3f\n" % scale)
    if display_size is not None:
        fh.write("display resize %d %d\n" % (display_size[0], display_size[1]))
    fh.write("display culling off\n")
    fh.write("animate goto start\n")
    if translation is not None:
        fh.write(
            "translate 1 %1.3f %1.3f %1.3f\n"
            % (translation[0], translation[1], transpation[2])
        )

    # Writing commands to loop through the trajectories
    # TODO: include frames and increment settings
    fh.write(
        "for {set frame 0} {$frame < %d} {incr frame %d} {\n"
        % (end_frame, increment_frame)
    )
    fh.write("    animate goto $frame\n")

    # Writing commands to reset opacity
    for n in range(ntrajs):
        fh.write("    set var%d [lindex $opa%d $frame]\n" % (n + 1, n + 1))
        fh.write("    material change opacity Transp%d $var%d\n" % (n + 1, n + 1))

    # Writing commands to render tga files
    fh.write('    set fname_tga snap.[format "%04d" $frame].tga\n')
    fh.write("    render TachyonInternal snapshots/$fname_tga\n")
    fh.write("    draw delete all\n")
    # TODO: custom sleep settings
    fh.write("    sleep %1.2f\n" % sleep)
    fh.write("}\n")


# TODO: add more options
def run_vmd_render():

    """Loads VMD and renders images, also composites into gif format"""

    os.system("mkdir snapshots")
    os.system("module load VMD/1.9.2")
    print("Running VMD")
    os.system("vmd -dispdev text -e vis_fms.vmd")


def get_mp4():
    print("Writing mp4")
    print("ffmpeg must be installed - not necessarily capable of running on fire!")
    os.system(
        "ffmpeg -r 4 -i snapshots/snap.%04d.tga -c:v libx264 -pix_fmt yuv420p -r 10 snap.mp4"
    )


def get_gif():

    print("Writing gif")
    os.system("convert -delay 20 -loop 1 snapshots/snap*tga snap.gif")



"""A repository for the representation and manipulation of AIMS-type trajectories."""

from . import iam, vis_aims  # noqa
from .blur import *  # noqa
from .bootstrap import *  # noqa
from .bundle import *  # noqa
from .fms90 import *  # noqa
from .geom import *  # noqa
from .pes import *  # noqa
from .plot import *  # noqa
from .pop import *  # noqa
from .rotate import *  # noqa
from .sine import *  # noqa
from .ued import *  # noqa
from .wrap import *  # noqa
from .xyz import *  # noqa

__version__ = "0.2.1"



import copy
import glob
import re

import matplotlib
import numpy as np

import aimsprop as ai

matplotlib.use("Agg")
import matplotlib.pyplot as plt


def build_traj(
    ICs,
):

    # => Parse/Align/Weight/Clean <= #

    # Parse a series of FMS90 trajectories
    main_dir = "/home/monikaw/chem/stilbene/4-aims"

    broken_frame = {}
    trajs = []
    labels = []
    for IC in ICs:
        input_dirs = [name for name in glob.glob("%s/%04d/*-run" % (main_dir, IC))]
        trajs_IC = []
        for run in input_dirs:
            mobj = re.match("\S+(\d+)-run", run)

        restart_index = np.array(
            [int(re.match("\S+(\d+)-run", run).group(1)) for run in input_dirs]
        )
        restart_inds = np.argsort(
            np.array(
                [int(re.match("\S+(\d+)-run", run).group(1)) for run in input_dirs]
            )
        )
        for ind in restart_inds:
            run = input_dirs[ind]
            ind1 = restart_index[ind]
            print(("Reading:", run))
            if (IC, ind1) in list(broken_frame.keys()):
                traj = ai.parse_fms90_dumpfile(
                    "%s" % run, scheme="mulliken", cutoff_time=broken_frame[(IC, ind1)]
                )
            else:
                traj = ai.parse_fms90_dumpfile("%s" % run, scheme="mulliken")
            trajs_IC.append(copy.copy(traj))
        print("merging restarts")
        traj_IC = ai.Bundle.merge(trajs_IC, [1.0] * len(trajs_IC), [IC] * len(trajs_IC))

        print("removing duplicate frames")
        traj_IC = traj_IC.remove_duplicates()

        print("interpolating across IC")
        ts = np.arange(0.0, max(traj_IC.ts), 400.0)
        traj_IC = traj_IC.interpolate_linear(ts)

        # Each dat contains np.array([[dE (eV), Osc, Tx, Ty, Tz]])
        dat = np.load("/home/monikaw/chem/stilbene/4-spectra/%04d/align.npz" % IC)[
            "dat"
        ]
        # Extract the transition dipole moments for alignment purposes
        tdm = dat[0, 2:]
        # Rotate each AIMS trajectory so the transition dipole moment vector is in aligned in z at t=0
        traj_IC = ai.rotate_frames_to_z(traj_IC, tdm)

        labels += [IC] * len(traj_IC.frames)
        trajs.append(copy.copy(traj_IC))

    # Parse the oscillator strength weights and transition dipole alignments
    # from the 4-spectra dir (computed by the align.py script)
    # Each dat contains np.array([[dE (eV), Osc, Tx, Ty, Tz]])
    dats = [
        np.load("/home/monikaw/chem/stilbene/4-spectra/%04d/align.npz" % x)["dat"]
        for x in ICs
    ]
    # Compute the normalized oscillator strength weights
    ws = np.array([dat[0, 1] for dat in dats])
    ws /= np.sum(ws)
    # Merge the trajectories into one super-big Bundle with uniform weights
    print("merging ICs")
    traj = ai.Bundle.merge(trajs, ws)
    print(("nframes", len(traj.frames)))

    return traj, trajs


if __name__ == "__main__":

    time_scale = 1.0 / 41.3413745758  # au to fs

    ICs = np.arange(100, 132, 1)
    traj, trajs = build_traj(ICs)
    ts = traj.ts

    # => UED Signal <= #

    print("calculating raw UED signal")
    R = np.linspace(0.0, 6.0, 200)
    traj = ai.compute_ued_simple(traj, "UED", R=R, alpha=8.0)

    # => Bootstrap Results <= #

    print("resampling trajectories")
    resampled_trajs = ai.bootstrap(traj, 1000, ICs)

    # => Plotting Population with Standard Deviation Error Bars <= #

    # Getting population statistics
    I1 = []
    I2 = []
    for ind, retraj in enumerate(resampled_trajs):
        pop1 = ai.pop.compute_population(retraj)
        I1.append(pop1[1])
        I2.append(pop1[2])
    std1 = np.std(I1, axis=0)
    std2 = np.std(I2, axis=0)
    avg_pop1 = np.average(I1, axis=0)
    avg_pop2 = np.average(I2, axis=0)

    # Plotting population with error bars
    plt.clf()
    plt.errorbar(ts * time_scale, avg_pop1, yerr=std1, fmt="none", ecolor="r")
    plt.errorbar(ts * time_scale, avg_pop2, yerr=std2, fmt="none", ecolor="b")
    plt.plot(ts * time_scale, avg_pop1, "r", label="$S_{0}$")
    plt.plot(ts * time_scale, avg_pop2, "b", label="$S_{1}$")

    plt.xlabel("t [fs]")
    plt.ylabel("Population")
    plt.axis([time_scale * min(ts), max(ts) * time_scale, -0.1, 1.1])
    plt.legend()
    plt.tight_layout()
    plt.savefig("Population.pdf")

    # => Plotting UED stats Heatmap <= #

    print("extracting UED statistics")
    avg, std = ai.extract_stats(resampled_trajs, "UED", diff=True)

    nlevel = 65

    plt.clf()
    vmax = np.max(np.abs(std))
    levels = np.linspace(0.0, vmax, nlevel)
    cticks = [0, +int(vmax)]
    plt.contourf(ts * time_scale, R, std.T, levels=levels)
    plt.colorbar(ticks=cticks)
    plt.xlabel("t [fs]")
    plt.ylabel("R[$\AA$]")
    plt.tight_layout()
    plt.savefig("ued_std.pdf")

    plt.clf()
    vmax = np.max(np.abs(avg))
    levels = np.linspace(-vmax, +vmax, nlevel)
    cticks = [-int(vmax), 0, +int(vmax)]
    plt.contourf(ts * time_scale, R, avg.T, cmap=plt.cm.bwr, levels=levels)
    plt.colorbar(ticks=cticks)
    plt.xlabel("t [fs]")
    plt.ylabel("R[$\AA$]")
    plt.tight_layout()
    plt.savefig("ued_avg.pdf")



import os
import re

import aimsprop.atom_data as atom_data
import aimsprop.manage_xyz as manage_xyz

# TODO: make more invariant to filename (sbatch.sh)

# run molpro calculations to obtain ionization potential information
def prep_molpro(
    opt_in,
):

    """Runs molpro calculations from template files and traj object

    Params:
        opt_in - [dict], options to override defaults

    """

    options = {
        # all paths must be absolute!
        "neutral_input": None,  # molpro input file for neutral
        "cation_input": None,  # molpro input file for cation
        "submit_input": None,  # sbatch file (runs neutral & cation in same job)
        "out_dir": None,  # dir for output
        "submit": False,  # whether to submit or just prep the job
        "cation_states": None,  # dict cation states corresponding to frame states
        "traj": None,  # traj object to run (all frames)
        "neutral_input2": "neutral.dat",  # modified molpro input file for neutral (must match sbatch.sh
        "cation_input2": "cation.dat",  # modified molpro input file for cation
        "submit_input2": "sbatch.sh",  # modified input script
        "njobs": 25,  # modified input script
    }

    for key, val in list(opt_in.items()):
        options[key] = val

    # TODO: throw error messages for invalid combination of options
    # override options
    neutral_input = options["neutral_input"]
    cation_input = options["cation_input"]
    submit_input = options["submit_input"]

    neutral_input2 = options["neutral_input2"]
    cation_input2 = options["cation_input2"]
    submit_input2 = options["submit_input2"]

    out_dir = options["out_dir"]
    cation_states = options["cation_states"]
    submit = options["submit"]
    traj = options["traj"]
    njobs = options["njobs"]

    frames_per_job = len(traj.frames) / njobs
    nframes = len(traj.frames)
    for job in range(njobs + 1):
        os.chdir(out_dir)
        os.system("mkdir %02djob" % job)
        os.system("cp %s %02djob" % (submit_input, job))
        os.system("cp run_molpro.py %02djob" % job)
        os.chdir("%02djob" % job)
        ind1 = frames_per_job * job
        ind2 = frames_per_job * (job + 1)
        if job == njobs:
            ind2 = nframes
        # update submission script
        update_sbatch("%d" % job, submit_input2, ind1=ind1, ind2=ind2)
        for ind, frame in enumerate(traj.frames[ind1:ind2]):
            # set up directory
            os.system("mkdir %04d" % (ind + ind1))
            os.chdir("%04d" % (ind + ind1))

            # modify neutral molpro params
            symbols = [atom_data.atom_symbol_table[N] for N in frame.N]
            insert_geom(frame.xyz, symbols, neutral_input, neutral_input2)
            update_state(frame.I, neutral_input2)
            write_xyz(frame, symbols, "geom.xyz")

            # modify cation molpro params
            symbols = [atom_data.atom_symbol_table[N] for N in frame.N]
            insert_geom(frame.xyz, symbols, cation_input, cation_input2)
            cI = cation_states[frame.I]
            update_state(cI, cation_input2)

            # create file to identify frame with job
            os.system(
                "echo 'labels: %4d %4d\nt: %12.6f' > %s"
                % (frame.label[0], frame.label[1], frame.t, "%04d_ID.txt" % ind)
            )

            # return to working directory
            os.chdir("../")

        # if requested, submit the job
        if submit == True:
            os.system("sbatch %s" % submit_input2)
        os.chdir("../")


def insert_geom(
    xyz,
    symbols,
    molpro_in,
    molpro_out,
):

    lines = open(molpro_in).readlines()
    fh = open(molpro_out, "w")
    for line in lines:
        mobj = re.match("\s*geometry=\{\s*", line)
        fh.write(line)
        if mobj:
            for ind, atom in enumerate(xyz):
                fh.write(
                    "%6s %12.6f %12.6f %12.6f\n"
                    % (symbols[ind], atom[0], atom[1], atom[2])
                )


def update_state(
    I,
    molpro_in,
):

    # Note: molpro is 1 based in state indexing (1 = groundstate, 2 = first excited state)
    # Note: fms is also 1 based in state indexing
    os.system('sed -i "s/xstate/%s/g" %s' % (I, molpro_in))


def update_sbatch(
    new_txt,
    sbatch_in,
    ind1,
    ind2,
):

    os.system('sed -i "s/ID/%s/g" %s' % (new_txt, sbatch_in))
    os.system('sed -i "s/ind1/%s/g" %s' % (ind1, sbatch_in))
    os.system('sed -i "s/ind2/%s/g" %s' % (ind2, sbatch_in))


def write_xyz(
    frame,
    symbols,
    filename,
):

    geom = []
    for ind, atom in enumerate(frame.xyz):
        geom.append((symbols[ind], atom[0], atom[1], atom[2]))

    manage_xyz.write_xyz(filename, geom)



from .bundle import Bundle


def unwrap_property(
    bundle: Bundle,
    key: str,
    period: float,
    N: int = 1,
    match_all: bool = False,
):
    """Unwrap a periodic property to be maximally continuous through time for
        each label.

    Params:
        bundle: the Bundle object to compute the property for (modified in
            place)
        key: the name of the original property
        period: the periodicity of the property
        N: the number of images to try in +/- wings
        match_all: whether to wrap all labels to same sign
    Return:
        bundle: reference to the input Bundle object. The
            property key is overwritten with the unwrapped property.
    """

    if match_all:
        ref_frame = bundle.subset_by_label(bundle.labels[0]).frames[0]
    else:
        ref_frame = None

    for label in bundle.labels:
        frame_old = ref_frame
        for frame in bundle.subset_by_label(label).frames:
            if frame_old is None:  # Property has just started
                frame_old = frame
                continue
            prop_old = frame_old.properties[key]
            prop = frame.properties[key]
            vals = [prop + k * period for k in range(-N, +N + 1)]
            frame.properties[key] = min(vals, key=lambda x: abs(x - prop_old))
            frame_old = frame
    return bundle



import numpy as np


def get_dyson(out_dir, traj, njobs):

    # TODO: verify frame corresponds to molpro run based on ID

    frames_per_job = len(traj.frames) / njobs
    nframes = len(traj.frames)
    for job in range(njobs + 1):
        ind1 = frames_per_job * job
        ind2 = frames_per_job * (job + 1)
        if job == njobs:
            ind2 = nframes
        for ind, frame in enumerate(traj.frames[ind1:ind2]):
            data = np.loadtxt("%s%02djob/%04d/norms.out" % (out_dir, job, ind + ind1))
            dyson_norms = {}
            if len(data) == 3:
                dyson_norms = np.array(
                    [
                        [int(data[1]) + 1, data[2]],
                    ]
                )
                frame.properties["dyson_norms"] = dyson_norms
            else:
                dyson_norms = np.array([(int(d[1]) + 1, d[2]) for d in data])
            frame.properties["dyson_norms"] = dyson_norms

    return traj



import matplotlib
import numpy as np

matplotlib.use("Agg")
import matplotlib.pyplot as plt
from matplotlib.figure import Figure

from . import pop  # for pop.compute_population
from .bundle import Bundle


def plot_scalar(
    filename: str,
    bundle: Bundle,
    key: str,
    ylabel: str = None,
    time_units: str = "au",
    legend_loc=1,
    state_colors: list = None,
    plot_average: bool = True,
    clf: bool = True,
) -> Figure:
    """Plot an AIMS scalar property (e.g., a "spaghetti plot" for a bond distance).

    Params:
        filename: the output PDF file
        bundle: the Bundle to plot properties for
        key: the key of the property
        ylabel: the label of the y-axis (defaults to key)
        time_units: "au" or "fs"
        legend_loc (legend location): location indicator for legend
        state_colors: list of colors to use for state
            plotting. None defaults to interpolation on jet colormap. For two
            or three states, many people prefer ['r', 'b', 'g'] or similar.
        plot_average: Plot averages for each state and total?
        clf : clear plot or not?
    Returns:
        plt handle for further modification
        saves figure to filename
    """

    if time_units == "au":
        time_scale = 1.0
    elif time_units == "fs":
        time_scale = 1.0 / 41.3413745758  # TODO: standardize this
    else:
        raise ValueError("Unknown time units: %s" % time_units)

    if state_colors:
        colors = state_colors
    else:
        cmap = matplotlib.cm.get_cmap("jet")
        colors = [
            cmap(float(x) / (len(bundle.Is) - 1))
            for x in reversed(list(range(len(bundle.Is))))
        ]

    if clf:
        plt.clf()
    if plot_average:
        # Plot average
        plt.plot(
            time_scale * np.array(bundle.ts),
            bundle.extract_property(key),
            "-k",
            linewidth=3.0,
        )
        # Plot state averages
        for Iind, I in enumerate(bundle.Is):
            bundle2 = bundle.subset_by_I(I)
            color = colors[Iind]
            plt.plot(
                time_scale * np.array(bundle2.ts),
                bundle2.extract_property(key),
                "-",
                color=color,
                linewidth=2.0,
            )
    # Plot individual frames
    for Iind, I in enumerate(bundle.Is):
        bundle2 = bundle.subset_by_I(I)
        color = colors[Iind]
        for lind, label in enumerate(bundle2.labels):
            bundle3 = bundle2.subset_by_label(label)
            plt.plot(
                time_scale * np.array(bundle3.ts),
                bundle3.extract_property(key),
                "-",
                color=color,
                linewidth=1.0,
                label=("State=%d" % I if lind == 0 else None),
            )

    plt.xlabel("t [%s]" % time_units)
    plt.ylabel(ylabel if ylabel else key)
    plt.axis("tight")  # TODO: Does not seem to respect this
    plt.legend(loc=legend_loc)
    plt.savefig(filename)
    return plt


def plot_vector(
    filename: str,
    bundle: Bundle,
    key: str,
    y: np.ndarray,
    ylabel: str = None,
    time_units: str = "au",
    diff: bool = False,
    cmap=plt.cm.bwr,
    levels: np.ndarray = None,
    nlevel: int = 65,
    twosided: bool = True,
    clf: bool = True,
) -> Figure:
    """Plot an AIMS vector property (e.g., a heatmap of a UED or PES signal).

    Params:
        filename: the output PDF file
        bundle: the Bundle to plot properties for
        key: the key of the property
        y: the indices of the y axis (e.g., R or Q or something
            like it)
        ylabel: the label of the y-axis (defaults to key)
        time_units: "au" or "fs"
        diff: is this a difference property from t=0?
        cmap: the desired colormap
        levels: the explicitly desired contour levels (1st
            priority).
        nlevel: number of evenly spaced contour levels to
            saturate data (2nd priority).
        twosided: is the colormap two-sided (used only with nlevel)
        clf: clear plot or not?
    Returns:
        plt handle for further modification
        saves figure to filename
    """

    if time_units == "au":
        time_scale = 1.0
    elif time_units == "fs":
        time_scale = 1.0 / 41.3413745758  # TODO: standardize this
    else:
        raise ValueError("Unknown time units: %s" % time_units)

    # Extract data to plot
    ts = np.array(bundle.ts)
    ys = y
    Vs = bundle.extract_property(key)

    # Difference properties
    if diff:
        Vs -= np.outer(np.ones_like(ts), Vs[0, :])

    # Levels and color ticks
    if levels is None:
        vmax = np.max(np.abs(Vs))
        if twosided:
            levels = np.linspace(-vmax, +vmax, nlevel)
            cticks = [-int(vmax), 0, +int(vmax)]
        else:
            levels = np.linspace(0, +vmax, nlevel)
            cticks = [0, +int(vmax)]
    else:
        vmax = np.max(levels)
        if twosided:
            cticks = [-int(vmax), 0, +int(vmax)]
        else:
            cticks = [0, +int(vmax)]

    if clf:
        plt.clf()
    hs = plt.contourf(
        time_scale * ts, ys, Vs.T, levels=levels, cmap=cmap, extend="both"
    )
    for h in hs.collections:
        h.set_edgecolor("face")
    plt.colorbar(ticks=cticks)
    plt.xlabel("t [%s]" % time_units)
    plt.ylabel(ylabel if ylabel else key)
    plt.axis("tight")  # TODO: Does not seem to respect this
    plt.savefig(filename)
    return plt


def plot_population(
    filename: str,
    bundle: Bundle,
    bundles: list,
    time_units: str = "au",
    legend_loc="right",
    state_colors: list = None,
    plot_total: bool = True,
    clf: bool = True,
    tmax: float = None,
) -> Figure:
    """Plot the AIMS state populations.

    Params:
        filename: the output PDF file
        bundle: the Bundle to plot populations for
        bundles: a list of Bundle objects, one for each
            IC. This is used for the "spaghetti" plots.
        time_units: "au" or "fs"
        legend_loc (legend location): location indicator for legend
        state_colors (list of colors): list of colors to use for state
            plotting. None defaults to interpolation on jet colormap. For two
            or three states, many people prefer ['r', 'b', 'g'] or similar.
        plot_total: Plot total population?
        clf: clear plot or not?
    Returns:
         plt handle for further modification
         saves figure to filename
    """

    if time_units == "au":
        time_scale = 1.0
    elif time_units == "fs":
        time_scale = 1.0 / 41.3413745758  # TODO: standardize this
    else:
        raise ValueError("Unknown time units: %s" % time_units)

    if state_colors:
        colors = state_colors
    else:
        cmap = matplotlib.cm.get_cmap("jet")
        colors = [
            cmap(float(x) / (len(bundle.Is) - 1))
            for x in reversed(list(range(len(bundle.Is))))
        ]

    if clf:
        plt.clf()

    # Spaghetti plots from bundles
    Imap = {I: Iind for Iind, I in enumerate(bundle.Is)}
    for bundle2 in bundles:
        ts = np.array(bundle2.ts)
        for I in bundle2.Is:
            bundle3 = bundle2.subset_by_I(I)
            t3s = bundle3.ts
            w3s = [sum(frame.w for frame in bundle3.subset_by_t(t).frames) for t in t3s]
            plt.plot(time_scale * t3s, w3s, "-", color=colors[Imap[I]], linewidth=1.0)

    # State populations
    ts = np.array(bundle.ts)
    populations = pop.compute_population(bundle)
    for Iind, I in enumerate(sorted(populations.keys())):
        plt.plot(
            time_scale * ts,
            populations[I],
            "-",
            color=colors[Iind],
            linewidth=2.0,
            label="State=%d" % I,
        )

    # Total population (to check norm, state cutoffs, etc)
    if plot_total:
        total = np.zeros_like(ts)
        for population in list(populations.values()):
            total += population
        plt.plot(time_scale * ts, total, "-", color="k", linewidth=2.0, label="Total")

    if tmax is None:
        tmax = max(ts) * time_scale

    plt.xlabel("t [%s]" % time_units)
    plt.ylabel("Population [-]")
    plt.axis([time_scale * min(ts), tmax, -0.1, 1.1])
    plt.legend(loc=legend_loc)
    plt.tight_layout()
    plt.savefig(filename)

    return plt



from mpl_toolkits.mplot3d import axes3d
import matplotlib.pyplot as plt
import numpy as np
import lebedev  # E.g., the one in aimsprop/iam/lebedev.py

nlebedev = 302
nomega = 12

leb = lebedev.Lebedev.build(nlebedev)

phis = leb.phi
zetas = leb.theta
omegas = np.linspace(0.0, 2.0 * np.pi, nomega, endpoint=False)

xyz = np.array([0.0, 0.0, 1.0])
uvw = np.array([1.1, 0.0, 1.0])

xyzs = []
uvws = []
ws = []
for phi, zeta in zip(phis, zetas):
    for omega in omegas:
        R1 = np.array(
            [
                [np.cos(phi), -np.sin(phi), 0.0],
                [np.sin(phi), np.cos(phi), 0.0],
                [0.0, 0.0, 1.0],
            ]
        )
        R2 = np.array(
            [
                [np.cos(zeta), 0.0, -np.sin(zeta)],
                [0.0, 1.0, 0.0],
                [np.sin(zeta), 0.0, np.cos(zeta)],
            ]
        )
        R3 = np.array(
            [
                [np.cos(omega), -np.sin(omega), 0.0],
                [np.sin(omega), np.cos(omega), 0.0],
                [0.0, 0.0, 1.0],
            ]
        )
        R = np.dot(R1, np.dot(R2, R3))
        xyz1 = np.dot(R, xyz)
        uvw1 = np.dot(R, uvw)
        xyzs.append(xyz1)
        uvws.append(uvw1)
        ws.append(np.sin(zeta))

xyzs = np.array(xyzs)
uvws = np.array(uvws)
ws = np.array(ws)

uvws -= xyzs

x = xyzs[:, 0]
y = xyzs[:, 1]
z = xyzs[:, 2]
u = uvws[:, 0]
v = uvws[:, 1]
w = uvws[:, 2]

fig = plt.figure()
ax = fig.gca(projection="3d")

h = ax.quiver(x, y, z, u, v, w, length=0.1)
print((type(h)))

plt.axis("off")
plt.show()



from pathlib import Path

import numpy as np
import pytest

import aimsprop as ap


@pytest.fixture(scope="module")
def bundle():
    # 1.  Parse a series of FMS90 trajectories that Hayley has run for ethylene
    trajs = [
        ap.parse_fms90(
            Path(__file__).parent / "test_data" / f"000{x}", scheme="mulliken"
        )
        for x in [2, 3]
    ]
    # 2. Merge the trajectories into one super-big Bundle with uniform weights
    traj = ap.Bundle.merge(trajs, ws=[1.0 / len(trajs)] * len(trajs), labels=[2, 3])

    # 3. Interpolate bundle with ~1 fs intervals, removing adaptive timesteps from AIMS
    ts = np.arange(0.0, max(traj.ts), 40.0)
    traj = traj.interpolate_nearest(ts)
    yield traj



import numpy as np

from . import lebedev


def rotation_quadrature(
    nlebedev,
    nomega,
):

    """Generate a rotation quadrature for SO(3) via a direct product of
        Lebedev grids on S(2) and linear grids on L2(2*pi).

    Params:
        nlebedev (int) - a Lebedev grid number for the number of nodes in the
            solid angle.
        nomega (int) - the number of nodes in the omega angle.
    Returns:
        Rs (list of np.ndarray of shape (3,3)) - Rotation matrices representing
            the quadrature nodes.
        ws (list of float) - the quadrature weights.
    """

    leb = lebedev.Lebedev.build(nlebedev)

    omega = np.linspace(0.0, 2.0 * np.pi, nomega, endpoint=False)
    womega = np.ones_like(omega) / nomega

    Rs = []
    ws = []

    for omega2, womega2 in zip(omega, womega):
        # Rotate about z
        R1 = np.array(
            [
                [np.cos(omega2), +np.sin(omega2), 0.0],
                [-np.sin(omega2), np.cos(omega2), 0.0],
                [0.0, 0.0, 1.0],
            ]
        )
        for U in range(len(leb.x)):
            # Rotate down to Omega from z
            n = np.array((0.0 - leb.x[U], 0.0 - leb.y[U], 1.0 - leb.z[U]))
            if leb.z[U] != 1.0:
                n /= np.sqrt(np.sum(n * n))
            R2 = np.eye(3) - 2.0 * np.outer(n, n)  # Householder reflector
            if leb.z[U] != 1.0:
                R2 *= -1.0  # Parity fix
            Rs.append(np.dot(R1, R2))
            ws.append(
                leb.w[U] * womega2 / (4.0 * np.pi)
            )  # Lebedev grids are normalized to 4*pi

    return Rs, ws



import numpy as np

p = np.random.rand() * 2.0 * np.pi
z = np.random.rand() * np.pi
o = np.random.rand() * 2.0 * np.pi

R1 = np.array(
    [
        [np.cos(p), -np.sin(p), 0.0],
        [np.sin(p), np.cos(p), 0.0],
        [0.0, 0.0, 1.0],
    ]
)
R2 = np.array(
    [
        [np.cos(z), 0.0, -np.sin(z)],
        [0.0, 1.0, 0.0],
        [np.sin(z), 0.0, np.cos(z)],
    ]
)
R3 = np.array(
    [
        [np.cos(o), -np.sin(o), 0.0],
        [np.sin(o), np.cos(o), 0.0],
        [0.0, 0.0, 1.0],
    ]
)

R = np.dot(R1, np.dot(R2, R3))

S = np.array(
    [
        [
            np.cos(p) * np.cos(z) * np.cos(o) - np.sin(p) * np.sin(o),
            -np.cos(p) * np.cos(z) * np.sin(o) - np.sin(p) * np.cos(o),
            -np.cos(p) * np.sin(z),
        ],
        [
            np.sin(p) * np.cos(z) * np.cos(o) + np.cos(p) * np.sin(o),
            -np.sin(p) * np.cos(z) * np.sin(o) + np.cos(p) * np.cos(o),
            -np.sin(p) * np.sin(z),
        ],
        [np.sin(z) * np.cos(o), -np.sin(z) * np.sin(o), np.cos(z)],
    ]
)

print((R - S))
print((np.max(np.abs(R - S))))



import numpy as np

from .bundle import Bundle


def rotate_frames(
    bundle: Bundle,
    R: np.ndarray,
) -> Bundle:
    """Rotate xyz coordinates for all frames in bundle by rotation matrix R.

    Performs xyz = xyz * R

    Params:
        bundle: the Bundle object to rotate coordinates in
            place (modified in place).
        R: rotation or transformation matrix, shape (3,3).
    Return:
        bundle: reference to the input Bundle object. The
            xyz coordinates are overwritten with the transformed xyz
            coordinates.
    """

    for frame in bundle.frames:
        frame.xyz = np.dot(frame.xyz, R)
    return bundle


def rotate_frames_to_z(
    bundle: Bundle,
    vec: np.ndarray,
) -> Bundle:
    """Rotate xyz coordinates for all frames in bundle so that vec is
        rotated onto z. Useful to align molecules to z according to transition
        dipole moment vector.

    Params:
        bundle: the Bundle object to rotate coordinates in
            place (modified in place).
        vec: vector in current coordinates to rotate
            onto +z, shape (3,)
    Return:
        bundle: reference to the input Bundle object. The
            xyz coordinates are overwritten with the transformed xyz
            coordinates.
    """

    n = np.array(vec)
    n /= np.sqrt(np.sum(n ** 2))
    v = 0.5 * (n + np.array((0.0, 0.0, 1.0)))
    v /= np.sqrt(np.sum(v ** 2))
    R = np.eye(3)
    R -= 2.0 * np.outer(v, v)
    R = np.dot(R, -np.eye(3))  # preserve chirality
    return rotate_frames(bundle, R)



import numpy as np

from .bundle import Bundle


def compute_pes(
    bundle: Bundle,
    carrier_frequency: float,
    alpha: float,
    eKT: np.ndarray,
) -> Bundle:

    """Compute the simple photoelectron spectroscopy, with Guassian blurring

    User is responsible for calculating and assigning properties to the bundle frames:
            Dyson Orbitals
            Ionization Potential (IP)

    Params:
        bundle: the Bundle object to compute the property for (modified in
            place)
        carrier_frequency: experimental probe pulse carrier frequency (hbar*omega)
        alpha: the Guassian blurring exponent
        eKT: electron energies

    Return:
        bundle: reference to the input Bundle object. The property
            key "pes" is set to computed PES property.
    """

    for frame in bundle.frames:
        IPs = frame.properties["IP"]
        dyson_norms = frame.properties["dyson_norms"]
        pes = np.zeros_like(eKT)
        for ind, (state, IP) in enumerate(IPs):
            dyson_norm = dyson_norms[np.where(dyson_norms[:, 0] == state), 1][0]
            pes += (
                dyson_norm
                * np.sqrt(alpha / np.pi)
                * np.exp(-alpha * (carrier_frequency - IP - eKT) ** 2)
            )
        frame.properties["pes"] = pes

    return bundle



def render():

    """Runs vmd and renders images

    Params:

    Returns:

    """

    # Requires:
    # Names of vmd and tcl scripts
    # Options:
    # render without opening vmd
    # open vmd before sourcing file
    raise NotImplementedError()



from typing import List

import numpy as np


class Frame(object):
    """Class Frame represents one "frame," a molecular geometry with associated time and label."""

    def __init__(
        self,
        label,  # TODO type?
        t: float,
        w: float,
        I: int,
        N: list,
        xyz: np.ndarray,
        widths: List[float],
        properties={},
    ):
        """Verbatim constructor.

        Arguments:
            label: label identifying basis function [e.g., TBF
                index (int), TBF pair index (int, int), etc]
            t: simulation time in au.
            w: weight of frame (e.g., Mulliken population, bundle
                weight, etc)
            I: electronic state label
            N: list of atomic numbers of molecule
            xyz: coordinates of nuclei in
                Angstrom, shape (natom,3)
            properties: dictionary mapping key (str) to user-defined
                numerical properties (float, complex, or np.ndarray)
        """

        self.label = label
        self.t = t
        self.w = w
        self.I = I
        self.N = N
        self.xyz = xyz
        self.properties = properties.copy()
        self.widths = widths

    def copy(self):
        """Make a copy of self that is sufficiently deep to prevent
        modification of self by modification of the copy's attributes or
        properties keys / references (property values are not deep copied)."""
        return Frame(
            label=self.label,
            t=self.t,
            w=self.w,
            I=self.I,
            N=self.N,
            xyz=self.xyz,
            widths=self.widths,
            properties=self.properties,  # __init__ makes a copy of this
        )

    def __lt__(
        self,
        other,
    ):
        """Replacement for python2 __cmp__, to help with sorting : Comparator basis is (t, label, I).

        Old __cmp__: return cmp((self.t, self.label, self.I), (other.t, other.label, other.I))

        Details: https://stackoverflow.com/questions/8276983/why-cant-i-use-the-method-cmp-in-python-3-as-for-python-2
        """
        return (self.t, self.label, self.I) < (other.t, other.label, other.I)

    def __str__(self):
        """Return frame properties"""
        return f"""label: {self.label}
        time: {self.t}
        weight: {self.w}
        state: {self.I}
        atomic number: {self.N}
        xyz: {self.xyz}
        widths: {self.widths}
        properties: {self.properties}
        """

    def _repr_html_(self):
        """Print nicely"""
        return "<p>" + str(self).replace("\n", "<br>") + "</p>"


class Bundle(object):
    """Class Bundle represents a list of Frames, with utility methods to
    extract and merge Bundles.

    Note that many methods below (e.g., subset_by_*) return shallow copies
    or "views" of the current Bundle object's frames. If a deeper copy
    is needed, one can easily call Bundle.copy(), which relies on
    Frame.copy().
    """

    def __init__(
        self,
        frames: list,  # TODO list of frames
    ):
        """Verbatim constructor.

        Params:
            frames: list of Frames in this Bundle.
        """

        self.frames = frames

    @property
    def labels(self):
        """Return all unique labels in this bundle, in sorted order."""
        return list(sorted(set([x.label for x in self.frames])))

    @property
    def ts(self):
        """Return all unique times in this bundle, in sorted order."""
        return np.array(sorted(set([x.t for x in self.frames])))

    @property
    def Is(self):
        """return all unique Is in this bundle, in sorted order."""
        return list(sorted(set([x.I for x in self.frames])))

    def copy(self):
        """Return a new Bundle with frames copied according to Frame.copy()."""
        return Bundle([frame.copy() for frame in self.frames])

    def subset_by_label(
        self,
        label,
    ):
        """Return a subset of this bundle containing all frames with a given label (Frame-sorted) (view)"""
        return Bundle(list(sorted([x for x in self.frames if x.label == label])))

    def subset_by_sublabel(
        self,
        label,
        index,
    ):
        """Return a subset of this bundle containing all frames with a given partial label (Frame-sorted) (view)"""
        return Bundle(list(sorted([x for x in self.frames if x.label[index] == label])))

    def subset_by_t(
        self,
        t,
        delta=1.0e-11,
    ):
        """Return a subset of this bundle containing all frames with a given time (Frame-sorted) (view).

        Note that due to possible weirdness with ULP errors in float t
        values, we grab all times within delta absolute error of t
        """
        return Bundle(list(sorted([x for x in self.frames if abs(x.t - t) < delta])))

    def subset_by_I(
        self,
        I,
    ):
        """Return a subset of this bundle containing all frames with a given I (Frame-sorted) (view)"""
        return Bundle(list(sorted([x for x in self.frames if x.I == I])))

    def __add__(
        self,
        other,
    ):
        """Concatenation operator to merge two bundles (view-based)"""
        return Bundle(self.frames + other.frames)

    def __mul__(
        self,
        w,
    ):
        """Return a new Bundle with all Frame objects multiplied by
        weight (new copy). This is useful to provide a weight on the initial
        condition due to e.g., oscillator strength and/or conformational population.
        """
        frames = []
        for frame in self.frames:
            frame2 = frame.copy()
            frame2.w *= w
            frames.append(frame2)
        return Bundle(frames)

    __rmul__ = __mul__

    @staticmethod
    def merge(
        bundles,
        ws,
        labels=None,
    ):
        """Merge a list of bundles together, with weighting factors (new copy).

        Params:
            bundles (list of Bundle): bundles to merge

            ws (list of float): weight factors of bundles (e.g., from
                oscillator strength or conformational well).
            labels (list of label): if provided, labels are updated to (label,
                frame.label). This is used, e.g., to label the frame by IC as
                well as by any bundle-specific labeling.
        Returns:
            Bundle: a single merged Bundle with updated weights (and labels)
        """

        frames = []
        for I, bundle in enumerate(bundles):
            for frame in bundle.frames:
                frame2 = frame.copy()
                frame2.w *= ws[I]
                if labels:
                    frame2.label = (labels[I], frame2.label)
                frames.append(frame2)
        return Bundle(frames)

    def interpolate_nearest(
        self,
        ts,
        delta=1.0e-11,
    ):
        """Return a new bundle with frame objects interpolated by nearest
            neighbor interpolation if t is inside the range of times of self's
            frames for each label (e.g., no extrapolation is performed). (new
            copy).

            Note that due to possible weirdness with ULP errors in float t
            values, we grab all times within delta absolute error of t

        Params:
            ts (list of float): times to interpolated new Bundle to
        Returns:
            Bundle: new bundle with interpolated frames.
        """

        frames = []
        for label in self.labels:
            bundle2 = self.subset_by_label(label)
            t2s = bundle2.ts
            for t in ts:
                if t < min(t2s - delta) or t > max(t2s + delta):
                    continue  # Out of range (no extrapolation)
                t2 = min(t2s, key=lambda x: abs(x - t))  # Closest value in t2s
                frames2 = bundle2.subset_by_t(t2).copy().frames
                for frame2 in frames2:
                    frame2.t = t
                frames += frames2
        return Bundle(list(sorted(frames)))

    def interpolate_linear(
        self,
        ts,
    ):
        """Return a new bundle with frame objects interpolated by linear
            interpolation if t is inside the range of times of self's
            frames for each label (e.g., no extrapolation is performed). (new
            copy).

            Using:
                f(t) = f(t0) + [ (f(t1) - f(t0)) / (t1 - t0) ] * (t - t0)
            Note:
                Not suitable for properties with large second derivative

        Arguments:
            ts (list of float): times to interpolated new Bundle to
        Returns:
            Bundle: new bundle with interpolated frames.
        """

        frames = []
        for label in self.labels:
            bundle2 = self.subset_by_label(label)
            t2s = bundle2.ts
            for t in ts:
                if t < min(t2s) or t >= max(t2s):
                    continue  # Out of range (no extrapolation)
                # Get points in time to linearly interpolate between
                t3 = t2s[np.where(t2s <= t)[0][-1]]
                t4 = t2s[np.where(t2s > t)[0][0]]
                # Retrieve frames at points of linear interpolation for all labels
                framei = bundle2.subset_by_t(t3).copy().frames
                framej = bundle2.subset_by_t(t4).copy().frames

                # TODO: Instead of this eliminate duplicate frames at beginning of method?
                if len(framei) >= 1:
                    # print "Warning: duplicate frames present!"
                    framei = framei[0]
                if len(framej) >= 1:
                    # print "Warning: duplicate frames present!"
                    framej = framej[0]

                # linearly interpolate weight
                w = framei.w + ((framej.w - framei.w) / (t4 - t3)) * (t - t3)
                # linearly interpolate position
                xyz = framei.xyz + ((framej.xyz - framei.xyz) / (t4 - t3)) * (t - t3)

                # TODO: create separate option for interpolation of properties (in case different interpolation scheme preferred?
                # also if property is of incompatible type (such as string)

                # linearly interpolate all other properties
                properties = {}
                for key, vali in list(framei.properties.items()):
                    valj = framej.properties[key]
                    valn = vali + ((valj - vali) / (t4 - t3)) * (t - t3)
                    properties[key] = valn

                # create updated frame at new time instace
                nframe = Frame(
                    label=label,
                    t=t,
                    w=w,
                    I=framei.I,
                    N=framei.N,
                    xyz=xyz,
                    properties=properties,
                    widths=framei.widths,
                )
                frames.append(nframe)

        return Bundle(list(sorted(frames)))

    def extract_property(
        self,
        key: str,
        normalize: bool = True,
        diff: bool = False,
    ):
        """Return a numpy array containing the time-history of a property,
            averaged over all Frames at each time (with frame weight applied).

        Params:
            key: the property key
            normalize: normalize the property by the sum of weights in
                each time?
            diff: difference property (zeroth-time value subtracted)?
        Returns:
            np.ndarray of shape (ntime, sizeof(prop)): the property
            expectation value. Time is always on the rows. If the property is
            scalar, this array will have ndim = 1. If the property is vector,
            this array with have ndim = 2. And so forth.
        """

        Vs = []
        for t in self.ts:
            bundle = self.subset_by_t(t)
            V = np.zeros_like(bundle.frames[0].properties[key])
            W = 0.0
            for frame in bundle.frames:
                V += frame.w * frame.properties[key]
                W += frame.w
            if normalize:
                V /= W
            Vs.append(V)
        if diff:
            R = np.copy(Vs[0])
            for V in Vs:
                V -= R
        return np.array(Vs)

    def remove_duplicates(
        self,
    ):
        """Return a new bundle with duplicate frame objects removed
            based on frame label and time index

        Returns:
            Bundle: new bundle with interpolated frames.
        """

        frames = []
        for ind1, frame1 in enumerate(self.frames):
            unique = True
            for ind2, frame2 in enumerate(self.frames[ind1 + 1 :]):
                if frame1.label == frame2.label and frame1.t == frame2.t:
                    unique = False
            if unique:
                frames.append(frame1)

        return Bundle(frames)



"""
Written by Alice Walker and Ruibin Liang

This script converts from Amber restart format to AIMS Geometry.dat files. It assumes your restart file has velocities.

To run, just give it the paths to your prmtop, restart data and desired output data as below.

main('test_data/parm.prmtop','test_data/','test_data/outs/')

"""

import argparse
from pathlib import Path

import numpy as np


def read_prmtop(prmpath: str):

    """This function parses a prmtop to extract atomic types/names and masses.

    Params:
      prmpath: the path to the prmtop file

    Returns:
       namelist: the atom names as a vector of strings.
       atommass: atomic masses as a vector of floats.

    """

    namelist = []
    atommass = []
    with open(prmpath) as f:
        for line in f:
            if "ATOM_NAME" in line:
                line = next(f)
                for line in f:
                    if "CHARGE" in line:
                        break
                    else:
                        line = line.strip("\n")
                        for c in range(0, len(line), 4):
                            namelist = namelist + [line[c : c + 3]]
            if "MASS" in line:
                line = next(f)
                for line in f:
                    if "ATOM_TYPE_INDEX" in line:
                        break
                    else:
                        a = line.split()
                        for x in a:
                            atommass.append(float(x))
    return namelist, atommass


def read_rst(rstpath: str):

    """This function parses a restart file to extract coordinates and velocities.

    Params:
      rstpath: the path to the restart file(s) to be processed. Can handle instances in subdirectories.

    Returns:
      atomcoord: the xyz coordinates as an N by 3 array.
      atomvel: the velocities as an N by 3 array.

    """

    atomcoord = np.array([])
    atomvel = np.array([])
    atomnum = 0
    with rstpath.open() as f:
        f.readline()
        line = f.readline()
        atomnum = int(line.split()[0])
        atomcoord = np.zeros((atomnum, 3))
        atomvel = np.zeros((atomnum, 3))
        for j in range(0, int((atomnum + 1) / 2)):
            line = f.readline()
            a = line.split()
            atomcoord[j * 2] = np.array([float(a[0]), float(a[1]), float(a[2])])
            if len(a) == 6:
                atomcoord[j * 2 + 1] = np.array([float(a[3]), float(a[4]), float(a[5])])
        for j in range(0, int((atomnum + 1) / 2)):
            line = f.readline()
            a = line.split()
            atomvel[j * 2] = np.array([float(a[0]), float(a[1]), float(a[2])])
            if len(a) == 6:
                atomvel[j * 2 + 1] = np.array([float(a[3]), float(a[4]), float(a[5])])
    return atomcoord, atomvel, atomnum


def output_geom(
    namelist: np.array,
    atommass: np.array,
    atomcoord: np.array,
    atomvel: np.array,
    atomnum: int,
    outpath: str,
    filename: str,
):

    """This function takes output from read_rst and read_prmtop as input and creates a Geometry.dat file, with appropriate numerical conversions to Bohr coordinates and a.u. time, to be used for AIMS simulations.

    Params:
      namelist: the vector of atom names as strings
      atommass: the vector of atomic masses as floats
      atomcoord: the 2d N by 3 array of atomic xyz coordinates
      atomvel: the 2d N by 3 array of atomic velocities
      atomnum: the total number of atoms in the system
      outpath: the directory to put all output files/directories
      filename: the name of the specific restart file to be converted

    Returns:
      A Geometry.dat file in the path outpath/filename-Geometry.dat

    """

    # Initialize new output directory if it doesn't exist

    if not outpath.exists():
        Path.mkdir(outpath)

    # Create directory structure from original files

    newdir = filename.parent
    finalpath = outpath.joinpath(newdir.relative_to(newdir.parent))
    Path.mkdir(finalpath, parents=True, exist_ok=True)

    with open(finalpath.joinpath("Geometry.dat"), "w+") as f:
        f.write("UNITS=BOHR\n")
        f.write(str(atomnum) + "\n")
        for j in range(0, atomnum):
            if namelist[j] == "Cl-":
                f.write(
                    "{0:<10s}".format(namelist[j][0:2])
                    + "".join(
                        "{0:>22.16f}".format(x / 0.52917724924) for x in atomcoord[j]
                    )
                    + "\n"
                )
            else:
                f.write(
                    "{0:<10s}".format(namelist[j][0])
                    + "".join(
                        "{0:>22.16f}".format(x / 0.52917724924) for x in atomcoord[j]
                    )
                    + "\n"
                )
        f.write("# momenta\n")
        for j in range(0, atomnum):
            for x in atomvel[j]:
                p = x * atommass[j] * 9.3499611711905144e-04 * 1.8228884855409500e03
                f.write(
                    "{0:>22.16f}".format(p)
                )  # time unit in rst7: 1/20.455 ps   length unit in rst7: angstrom    mass unit in amber: atomic mass unit
            f.write("\n")


def convert_geometries(
    prmpath: str,
    rstpath: str,
    outpath: str,
):

    """This function creates a directory structure and output to convert a set of geometries from Amber prmtop/rst format to FMS90 Geometry.dat format.

    Params:
      prmpath: the path to the prmtop file
      rstpath: the path to the restart file
      outpath: the path to the output files

    Returns:
      All converted geometries in the path outpath/filename.
    """

    findme = "*/*.rst*"

    Prmpath = Path(prmpath)
    Rstpath = Path(rstpath)
    Outpath = Path(outpath)

    # Get initial information from parmtop
    names, masses = read_prmtop(Prmpath)

    # Find all restart files in test data
    filenames = list(Rstpath.glob(findme))

    # Create new dats
    for filename in filenames:
        coords, vel, numats = read_rst(filename)
        output_geom(names, masses, coords, vel, numats, Outpath, filename)


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "-p", "--prm", dest="prmpath", help="/path/to/input/prmtop", required=True
    )
    parser.add_argument(
        "-r", "--rst", dest="rstpath", help="/path/to/input/restarts", required=True
    )
    parser.add_argument(
        "-o", "--out", dest="outpath", help="/path/to/outputs/", required=True
    )
    args = parser.parse_args()
    convert_geometries(args.prmpath, args.rstpath, args.outpath)



import copy

import numpy as np

from . import bundle


def bootstrap(
    input_bundle: bundle.Bundle,
    nsamples: int,
    ICs: list,
    label_ind: int = 0,
):
    """Resample Bundle Subset According to Bootstrap Algorithm

    Params:
        input_bundle: the Bundle object to resample with replacement
        nsamples: number of resampling sets to conduct
        label_ind: sub label index from which to sample (IC = 0)
    Returns:
        resampled_bundles (list of Bundles [nsamples]):  re-weighted list of bundles

    Note:
        Removed parameters: sample_labels - list of bundle labels from which to resample
    """

    # accumulate samples
    labels = input_bundle.labels
    nICs = len(ICs)
    resampled_input_bundles = []
    for ind in range(nsamples):

        # copy of bundle to modify
        bundle1 = copy.copy(input_bundle)

        # generate random numbers for resampling
        samples_inds = np.random.randint(low=0, high=nICs, size=nICs)

        # get new sample set
        samples = []
        new_labels = []
        for ind1, sind in enumerate(samples_inds):
            IC = ICs[sind]
            [samples.append(label) for label in labels if label[label_ind] == IC]
            [new_labels.append(ind1) for label in labels if label[label_ind] == IC]

        # separate new sample set from input_bundle
        ws = 0.0
        input_bundles = [bundle1.subset_by_label(sample) for sample in samples]
        t0 = bundle1.ts[0]
        # re-weight bundles
        for input_bundle_t in input_bundles:
            input_bundles_t0 = input_bundle_t.subset_by_t(t0)
            for frame in input_bundles_t0.frames:
                ws += frame.w
        ws = 1.0 / ws

        # merge input_bundles
        bundle1 = bundle.Bundle.merge(
            input_bundles, [ws] * len(bundle1.frames), labels=new_labels
        )
        resampled_input_bundles.append(bundle1)
        print((ind + 1, "/", nsamples, "complete"))

    return resampled_input_bundles


def extract_stats(
    bundles: list,
    key: str,
    diff=False,
):
    """Extract Standard Deviation and Average of Property from Set of Resampled Bundles

    Params:
        bundles: re-weighted list of bundles
        key: the property key
        diff: difference property (zeroth-time value subtracted)?
    Returns:
        avg (np.ndarray of shape (ntime, sizeof(prop))): the average property expectation value.
        std (np.ndarray of shape (ntime, sizeof(prop))): the standard deviation of the property expectation value.

    """

    props = []
    for ind, bundle in enumerate(bundles):
        PROP = bundle.extract_property(key)
        if diff == True:
            PROP -= np.outer(np.ones((PROP.shape[0],)), PROP[0, :])
        props.append(PROP)

    std = np.std(props, axis=0)
    avg = np.average(props, axis=0)

    return avg, std



import numpy as np
import pytest as pt

import aimsprop as ai


def test_ued_simple(bundle):

    # 1. Compute the UED spectrum.
    #    Test if the computed UED spectrum at time 0 is correct.
    R = np.linspace(0.0, 6.0, 200)
    bundle = ai.compute_ued_simple(bundle, "UED", R=R, alpha=8.0)
    assert bundle.frames[0].properties["UED"][0] == pt.approx(8.95739037e-07, rel=1e-7)



import numpy as np

from . import formfactor, rotation


def compute_diffraction(
    bundle,
    key,
    s,
    eta,
    L,
    nlebedev=74,
    nomega=12,
    mode="xray",
    form="raw",
    anisotropy="cos2",
    print_level=False,
):

    """Compute the I(s, eta) elastic scattering signal for a Bundle.
         See aimsprop/notes/ued for details on this property.

    Notes:
        * All frames for each initial condition (IC) in bundle should be aligned so
        that the transition dipole moment from S0 -> Sex at t=0 is on z. This
        is required for proper computation of anisotropy.
        * All frames should be weighted by geometric considerations at the IC
        (e.g., conformational wells, Wigner weights, etc), by the cross
        section for the optical transition at the IC (e.g., oscillator
        strength and excitation energy window), and by the frame weight due
        to non-adiabatic dynamics.

    Params:
        bundle (Bundle) - the Bundle object to compute the property for (modified in
            place)
        key (str) - the name of the property.
        s (np.ndarray) - list of scattering vector norms in Angstrom^-1. The
            relationship between s and theta (scattering angle) is given as,
                s = 4 pi / L * sin(theta / 2).
        eta (np.ndarray) - list of azimuthal scattering angles in radians.
        L (float) - effective wavelength of scattering particle (x-ray
            wavelength or UED deBroglie wavelength) in Angstrom. Used to
            convert through scattering angle theta.
        nlebedev (int) - Lebedev number to use for solid angle orientation
            quadrature.
        nomega (int) - number of uniform quadrature points to use for plane
            orientation quadrature.
        mode (str) - 'xray' or 'ued' for selection of form factors
        form (str) - 'raw' or 'mod' for modified/raw diffraction intensities
            I(s) or M(s).
        anisotropy (str) - 'none' or 'cos2' for isotropic of cos^2 (z)
            anisotropty.
        print_level (bool) - print progress if true (useful to track long
            property computations)
    Result/Return:
        bundle - reference to the input Bundle object. The properties
    """

    # Validity checks
    if mode not in ["xray", "ued"]:
        raise ValueError("Unknown mode: %s" % mode)
    if form not in ["raw", "mod"]:
        raise ValueError("Unknown form: %s" % form)
    if anisotropy not in ["none", "cos2"]:
        raise ValueError("Unknown anisotropy: %s" % anisotropy)

    # Compute scattering angles via Bragg equation
    theta = 2.0 * np.arcsin(s * L / (4.0 * np.pi))
    tt, ee = np.meshgrid(theta, eta, indexing="ij")
    ss, ee = np.meshgrid(s, eta, indexing="ij")
    # Compute scattering vectors
    sx = ss * np.cos(tt / 2.0) * np.sin(ee)
    sy = ss * np.sin(tt / 2.0)
    sz = ss * np.cos(tt / 2.0) * np.cos(ee)

    # Get a rotation quadrature for the orientations of the frames
    if nlebedev == 1 and nomega == 1:
        # Fixed orientation
        Rs = [np.eye(3)]
        ws = [1.0]
    else:
        # Rotation quadrature
        Rs, ws = rotation.rotation_quadrature(nlebedev=nlebedev, nomega=nomega)

    # Get atomic form factors for appropriate x-ray/ued mode
    factors = formfactor.AtomicFormFactor.build_factors(bundle.frames[0], mode=mode)

    # Compute atomic scattering Iat
    D = np.zeros_like(sx)
    for A, factor in enumerate(factors):
        F = factor.evaluate_N(qx=sx, qy=sy, qz=sz, x=0.0, y=0.0, z=0.0)
        D += (np.abs(F) ** 2).real

    # Compute IAM scattering, integrating over all orientation angles
    for find, frame in enumerate(bundle.frames):
        if print_level:
            print(("Frame %5d of %5d" % (find, len(bundle.frames))))
        I = np.zeros_like(sx)
        for R, w in zip(Rs, ws):
            # cos(z)^2 pump anisotropy
            cos2 = R[2, 2] ** 2 if anisotropy == "cos2" else 1.0
            # Rotated molecule
            xyz = np.dot(frame.xyz, R)
            # Compute diffraction
            N = np.zeros_like(I, dtype=complex)
            for A, factor in enumerate(factors):
                x = xyz[A, 0]
                y = xyz[A, 1]
                z = xyz[A, 2]
                N += factor.evaluate_N(qx=sx, qy=sy, qz=sz, x=x, y=y, z=z)
            F = (np.abs(N) ** 2).real
            if form == "mod":
                F = (F - D) / D
            I += w * cos2 * F
        frame.properties[key] = I

    return bundle


def compute_diffraction_fast(
    bundle,
    key,
    s,
    eta,
    L,
    nlebedev=74,
    nomega=12,
    mode="xray",
    form="raw",
    anisotropy="cos2",
    print_level=False,
):

    """Compute the I(s, eta) elastic scattering signal for a Bundle.
         See aimsprop/notes/ued for details on this property.

    Notes:
        * All frames for each initial condition (IC) in bundle should be aligned so
        that the transition dipole moment from S0 -> Sex at t=0 is on z. This
        is required for proper computation of anisotropy.
        * All frames should be weighted by geometric considerations at the IC
        (e.g., conformational wells, Wigner weights, etc), by the cross
        section for the optical transition at the IC (e.g., oscillator
        strength and excitation energy window), and by the frame weight due
        to non-adiabatic dynamics.

    Params:
        bundle (Bundle) - the Bundle object to compute the property for (modified in
            place)
        key (str) - the name of the property.
        s (np.ndarray) - list of scattering vector norms in Angstrom^-1. The
            relationship between s and theta (scattering angle) is given as,
                s = 4 pi / L * sin(theta / 2).
        eta (np.ndarray) - list of azimuthal scattering angles in radians.
        L (float) - effective wavelength of scattering particle (x-ray
            wavelength or UED deBroglie wavelength) in Angstrom. Used to
            convert through scattering angle theta.
        nlebedev (int) - Lebedev number to use for solid angle orientation
            quadrature.
        nomega (int) - number of uniform quadrature points to use for plane
            orientation quadrature.
        mode (str) - 'xray' or 'ued' for selection of form factors
        form (str) - 'raw' or 'mod' for modified/raw diffraction intensities
            I(s) or M(s).
        anisotropy (str) - 'none' or 'cos2' for isotropic of cos^2 (z)
            anisotropty.
        print_level (bool) - print progress if true (useful to track long
            property computations)
    Result/Return:
        bundle - reference to the input Bundle object. The properties
    """

    # Validity checks
    if mode not in ["xray", "ued"]:
        raise ValueError("Unknown mode: %s" % mode)
    if form not in ["raw", "mod"]:
        raise ValueError("Unknown form: %s" % form)
    if anisotropy not in ["none", "cos2"]:
        raise ValueError("Unknown anisotropy: %s" % anisotropy)

    # Get a rotation quadrature for the orientations of the frames
    if nlebedev == 1 and nomega == 1:
        # Fixed orientation
        Rs = [np.eye(3)]
        ws = [1.0]
    else:
        # Rotation quadrature
        Rs, ws = rotation.rotation_quadrature(nlebedev=nlebedev, nomega=nomega)

    # Get atomic form factors for appropriate x-ray/ued mode
    factors = formfactor.AtomicFormFactor.build_factors(bundle.frames[0], mode=mode)

    import lightspeed as ls

    from . import ext

    s2s = ls.Tensor.array(s)
    eta2s = ls.Tensor.array(eta)

    R2s = ls.Tensor.zeros((len(Rs), 3, 3))
    for (
        Rind,
        R,
    ) in enumerate(Rs):
        R2s[Rind, :, :] = R
    w2s = ls.Tensor.array(ws)

    fA = ls.Tensor.zeros((len(factors), s.size))
    for A, factor in enumerate(factors):
        fA[A, :] = factor.evaluate(qx=0.0, qy=0.0, qz=s)

    # Compute IAM scattering, integrating over all orientation angles
    for find, frame in enumerate(bundle.frames):
        if print_level:
            print(("Frame %5d of %5d" % (find, len(bundle.frames))))
        xyz = ls.Tensor.array(frame.xyz)
        I = ext.compute_diffraction(
            L,
            s2s,
            eta2s,
            xyz,
            fA,
            R2s,
            w2s,
            True if anisotropy == "cos2" else False,
            True if form == "mod" else False,
        )
        frame.properties[key] = np.array(I)

    return bundle


def compute_diffraction_moments_fast(
    bundle,
    key,
    s,
    L,
    nlebedev=74,
    nomega=12,
    mode="xray",
    form="raw",
    anisotropy="cos2",
    print_level=False,
):

    """Compute the I(s, eta) elastic scattering moments for a Bundle.
         See aimsprop/notes/ued for details on this property.

    Notes:
        * All frames for each initial condition (IC) in bundle should be aligned so
        that the transition dipole moment from S0 -> Sex at t=0 is on z. This
        is required for proper computation of anisotropy.
        * All frames should be weighted by geometric considerations at the IC
        (e.g., conformational wells, Wigner weights, etc), by the cross
        section for the optical transition at the IC (e.g., oscillator
        strength and excitation energy window), and by the frame weight due
        to non-adiabatic dynamics.

    Params:
        bundle (Bundle) - the Bundle object to compute the property for (modified in
            place)
        key (str) - the name of the property.
        s (np.ndarray) - list of scattering vector norms in Angstrom^-1. The
            relationship between s and theta (scattering angle) is given as,
                s = 4 pi / L * sin(theta / 2).
        L (float) - effective wavelength of scattering particle (x-ray
            wavelength or UED deBroglie wavelength) in Angstrom. Used to
            convert through scattering angle theta.
        nlebedev (int) - Lebedev number to use for solid angle orientation
            quadrature.
        nomega (int) - number of uniform quadrature points to use for plane
            orientation quadrature.
        mode (str) - 'xray' or 'ued' for selection of form factors
        form (str) - 'raw' or 'mod' for modified/raw diffraction intensities
            I(s) or M(s).
        anisotropy (str) - 'none' or 'cos2' for isotropic of cos^2 (z)
            anisotropty.
        print_level (bool) - print progress if true (useful to track long
            property computations)
    Result/Return:
        bundle - reference to the input Bundle object. The properties "key-0"
            and "key-2" are added to each frame of the Bundle.
    """

    # Validity checks
    if mode not in ["xray", "ued"]:
        raise ValueError("Unknown mode: %s" % mode)
    if form not in ["raw", "mod"]:
        raise ValueError("Unknown form: %s" % form)
    if anisotropy not in ["none", "cos2"]:
        raise ValueError("Unknown anisotropy: %s" % anisotropy)

    # Special angles to collocate
    eta = np.array([0.0, np.pi / 2.0])

    # Get a rotation quadrature for the orientations of the frames
    if nlebedev == 1 and nomega == 1:
        # Fixed orientation
        Rs = [np.eye(3)]
        ws = [1.0]
    else:
        # Rotation quadrature
        Rs, ws = rotation.rotation_quadrature(nlebedev=nlebedev, nomega=nomega)

    # Get atomic form factors for appropriate x-ray/ued mode
    factors = formfactor.AtomicFormFactor.build_factors(bundle.frames[0], mode=mode)

    import lightspeed as ls

    from . import ext

    s2s = ls.Tensor.array(s)
    eta2s = ls.Tensor.array(eta)

    R2s = ls.Tensor.zeros((len(Rs), 3, 3))
    for (
        Rind,
        R,
    ) in enumerate(Rs):
        R2s[Rind, :, :] = R
    w2s = ls.Tensor.array(ws)

    fA = ls.Tensor.zeros((len(factors), s.size))
    for A, factor in enumerate(factors):
        fA[A, :] = factor.evaluate(qx=0.0, qy=0.0, qz=s)

    # Compute IAM scattering, integrating over all orientation angles
    for find, frame in enumerate(bundle.frames):
        if print_level:
            print(("Frame %5d of %5d" % (find, len(bundle.frames))))
        xyz = ls.Tensor.array(frame.xyz)
        I = ext.compute_diffraction(
            L,
            s2s,
            eta2s,
            xyz,
            fA,
            R2s,
            w2s,
            True if anisotropy == "cos2" else False,
            True if form == "mod" else False,
        )
        # Moment computation
        I0 = 0.5 * (I[:, 0] + I[:, 1])
        I1 = 0.5 * (I[:, 0] - I[:, 1])
        frame.properties["%s-0" % key] = I0
        frame.properties["%s-2" % key] = I1

    return bundle


def compute_diffraction_from_moments(
    bundle,
    key,
    eta,
):

    for find, frame in enumerate(bundle.frames):
        I = np.outer(frame.properties["%s-0" % key], np.cos(0 * eta)) + np.outer(
            frame.properties["%s-2" % key], np.cos(2 * eta)
        )
        frame.properties[key] = I

    return bundle


def compute_diffraction_moments_analytical(
    bundle,
    key,
    s,
    L,
    mode="xray",
    form="raw",
    anisotropy="perpendicular",
    print_level=False,
):

    """Compute the I(s, eta) elastic scattering moments for a Bundle.
         See aimsprop/notes/ued for details on this property.

    Notes:
        * All frames for each initial condition (IC) in bundle should be aligned so
        that the transition dipole moment from S0 -> Sex at t=0 is on z. This
        is required for proper computation of anisotropy.
        * All frames should be weighted by geometric considerations at the IC
        (e.g., conformational wells, Wigner weights, etc), by the cross
        section for the optical transition at the IC (e.g., oscillator
        strength and excitation energy window), and by the frame weight due
        to non-adiabatic dynamics.

    Params:
        bundle (Bundle) - the Bundle object to compute the property for (modified in
            place)
        key (str) - the name of the property.
        s (np.ndarray) - list of scattering vector norms in Angstrom^-1. The
            relationship between s and theta (scattering angle) is given as,
                s = 4 pi / L * sin(theta / 2).
        L (float) - effective wavelength of scattering particle (x-ray
            wavelength or UED deBroglie wavelength) in Angstrom. Used to
            convert through scattering angle theta.
        nlebedev (int) - Lebedev number to use for solid angle orientation
            quadrature.
        nomega (int) - number of uniform quadrature points to use for plane
            orientation quadrature.
        mode (str) - 'xray' or 'ued' for selection of form factors
        form (str) - 'raw' or 'mod' for modified/raw diffraction intensities
            I(s) or M(s).
        anisotropy (str) - 'none' or 'perpendicular' or 'parallel'
        print_level (bool) - print progress if true (useful to track long
            property computations)
    Result/Return:
        bundle - reference to the input Bundle object. The properties "key-0"
            and "key-2" are added to each frame of the Bundle.
    """

    # Validity checks
    if mode not in ["xray", "ued"]:
        raise ValueError("Unknown mode: %s" % mode)
    if form not in ["raw", "mod"]:
        raise ValueError("Unknown form: %s" % form)
    if anisotropy not in ["none", "perpendicular", "parallel"]:
        raise ValueError("Unknown anisotropy: %s" % anisotropy)

    # Compute scattering angles via Bragg equation
    theta = 2.0 * np.arcsin(s * L / (4.0 * np.pi))

    # Get atomic form factors for appropriate x-ray/ued mode
    factors = formfactor.AtomicFormFactor.build_factors(bundle.frames[0], mode=mode)

    # Collocate the atomic form factors
    f = np.zeros((len(factors), s.size))
    for A, factor in enumerate(factors):
        f[A, :] = factor.evaluate(qx=0.0, qy=0.0, qz=s)

    # Compute atomic scattering Iat
    D = np.zeros_like(s)
    for A, factor in enumerate(factors):
        D += f[A, :] ** 2

    # Selection fraction
    F = 1.0 if anisotropy == "isotropic" else 1.0 / 3.0

    # Pairs to include
    ABpairs = []
    for A in range(bundle.frames[0].xyz.shape[0]):
        for B in range(bundle.frames[0].xyz.shape[0]):
            if A >= B:
                continue
            ABpairs.append((A, B))

    # Diffraction moment computation
    for find, frame in enumerate(bundle.frames):
        if print_level:
            print(("Frame %5d of %5d" % (find, len(bundle.frames))))
        # Geometry
        xyz = frame.xyz
        # Target
        I0 = np.zeros_like(s)
        I2 = np.zeros_like(s)
        for A, B in ABpairs:
            # Geometry
            d = xyz[A, :] - xyz[B, :]
            r2 = np.sum(d ** 2)
            r = np.sqrt(r2)
            sg2 = np.sum(d[:2] ** 2) / r2
            sr = s * r
            # Bessel functions
            J0 = np.sin(sr) / sr
            J0[sr == 0.0] = 1.0
            J1sr = np.sin(sr) / sr ** 3 - np.cos(sr) / sr ** 2
            J1sr[sr == 0.0] = 1.0 / 3.0
            J2 = (3.0 / sr ** 2 - 1.0) * np.sin(sr) / sr - 3.0 * np.cos(sr) / sr ** 2
            J2[sr == 0.0] = 0.0
            # Kernels
            if anisotropy == "isotropic":
                I0 += 2.0 * f[A, :] * f[B, :] * J0
            elif anisotropy == "perpendicular":
                Iz = (
                    2.0
                    * f[A, :]
                    * f[B, :]
                    * (
                        J1sr
                        - (sg2 + (2.0 - 3.0 * sg2) * np.cos(0.5 * theta) ** 2)
                        * J2
                        / 2.0
                    )
                )
                Ix = 2.0 * f[A, :] * f[B, :] * (J1sr - (sg2) * J2 / 2.0)
                I0 += 0.5 * (Iz + Ix)
                I2 += 0.5 * (Iz - Ix)
            elif anisotropy == "parallel":
                I0 += (
                    2.0
                    * f[A, :]
                    * f[B, :]
                    * (
                        J1sr
                        - (sg2 + (2.0 - 3.0 * sg2) * np.sin(0.5 * theta) ** 2)
                        * J2
                        / 2.0
                    )
                )
        # Modified detector pattern
        if form == "raw":
            I0 += F * D
        if form == "mod":
            I0 /= D
            I2 /= D
        # Placement
        frame.properties["%s-0" % (key)] = I0
        frame.properties["%s-2" % (key)] = I2

    return bundle


# TODO: These are deprecated, as they are not fully correct for elastic scattering
# def compute_diffraction_moments(
#     bundle,
#     key,
#     q,
#     factors,
#     nlebedev,
#     nlebedev2,
#     nomega2,
#     nlegendre=2,
#     print_level=False,
#     ):
#
#     """ Compute the IAM X-Ray Diffraction or UED moments property. See
#         aimsprop/notes/xray for details on these moments.
#
#     Notes:
#         * All frames for each initial condition (IC) in bundle should be aligned so
#         that the transition dipole moment from S0 -> Sex at t=0 is on z. This
#         is required for proper computation of I2 (I0 is invariant to this).
#         * All frames should be weighted by geometric considerations at the IC
#         (e.g., conformational wells, Wigner weights, etc), by the cross
#         section for the optical transition at the IC (e.g., oscillator
#         strength and excitation energy window), and by the frame weight due
#         to non-adiabatic dynamics.
#
#     Params:
#         bundle (Bundle) - the Bundle object to compute the property for (modified in
#             place)
#         key (str) - the name of the property.
#         q (np.ndarray) - the 1d array of |q| values to collocate the
#             diffraction signal to.
#         factors (list of AtomicFormFactor) - the list of AtomicFormFactor
#             objects. The choice of xray/ued is made by the "mode" field of each
#             AtomicFormFactor.
#         nlebedev (int) - a Lebedev number for the number of grid points to use
#             for angular integration of the Legendre coefficients.
#         nlebedev2 (int) - a Lebedev number for the number of grid points to use
#             for angular sampling in the Omega angle (surface angle).
#         nomega2 (int) - the number of grid points to use for angular sampling
#             in the omega angle (surface orientation).
#         nlegendre (even int) - the maximum Legendre polynomial order to
#             evaluate (usually 0 and 2 are the only Legendre polynomial
#             coefficients with any signal).
#         print_level (bool) - print progress if true (useful to track long
#             property computations)
#     Result/Return:
#         bundle - reference to the input Bundle object. The properties
#             key-l are set where l is [0, 2, ..., nlegendre].
#     """
#     if nlegendre % 2: raise ValueError('Can only ask for even Legendre functions')
#
#     # Lebedev grid for S(\vec q)
#     leb = lebedev.Lebedev.build(nlebedev)
#     # Direct product grid for \vec q
#     qx = np.outer(q, leb.x)
#     qy = np.outer(q, leb.y)
#     qz = np.outer(q, leb.z)
#
#     # Rotation quadrature
#     Rs, ws = rotation.rotation_quadrature(nomega=nomega2, nlebedev=nlebedev2)
#
#     for find, frame in enumerate(bundle.frames):
#         if print_level:
#             print 'Frame %5d of %5d' % (find, len(bundle.frames))
#         # Compute N(\vec q) = \sum_{A} f_A (\vec q) * \exp(-1.j * \vec q * \vec r)
#         N = np.zeros((len(q), len(leb.x)), dtype=complex)
#         for A, factor in enumerate(factors):
#             x = frame.xyz[A,0]
#             y = frame.xyz[A,1]
#             z = frame.xyz[A,2]
#             N += factor.evaluate_N(qx=qx,qy=qy,qz=qz,x=x,y=y,z=z)
#         # Compute I(\vec q) = N(\vec q)**2 for this frame
#         I = (np.abs(N)**2).real
#
#         # Do angle integration
#         Ils = { l: np.zeros((len(q),)) for l in range(0, nlegendre+1, 2) }
#         qxyz = leb.xyz
#         for R2, w2 in zip(Rs, ws):
#             # Account for cos(z)^2 weight
#             cos2 = np.sum(R2[:,2])**2
#             # Rotate the lebedev grid
#             qxyz2 = np.dot(qxyz, R2)
#             qx2 = qxyz2[:,0]
#             qy2 = qxyz2[:,1]
#             qz2 = qxyz2[:,2]
#             # Weight by zonal harmonics (evaluated in rotated grid)
#             Y = legendre.zonal2(qz2, nlegendre)
#             for l in range(0, nlegendre+1, 2):
#                 Ils[l] += w2 * cos2 * np.einsum('qw,w->q', I, leb.w * Y[l/2, :])
#
#         # Assign the properties to frame
#         for l in range(0, nlegendre+1, 2):
#             frame.properties['%s-%d' % (key, l)] = Ils[l]
#
#     return bundle
#
# def compute_diffraction_moment0(
#     bundle,
#     key,
#     q,
#     factors,
#     nlebedev,
#     print_level=False,
#     ):
#
#     """ Compute the IAM X-Ray Diffraction or UED moment property for only l=0
#         (faster due to lack of rotation quadratures). See aimsprop/notes/xray for
#         details on this moment
#
#     Notes:
#         * This moment is invariant to the orientation of the frames, so alignment to the IC
#         * All frames should be weighted by geometric considerations at the IC
#         (e.g., conformational wells, Wigner weights, etc), by the cross
#         section for the optical transition at the IC (e.g., oscillator
#         strength and excitation energy window), and by the frame weight due
#         to non-adiabatic dynamics.
#
#     Params:
#         bundle (Bundle) - the Bundle object to compute the property for (modified in
#             place)
#         key (str) - the name of the property.
#         q (np.ndarray) - the 1d array of |q| values to collocate the
#             diffraction signal to.
#         factors (list of AtomicFormFactor) - the list of AtomicFormFactor
#             objects. The choice of xray/ued is made by the "mode" field of each
#             AtomicFormFactor.
#         nlebedev (int) - a Lebedev number for the number of grid points to use
#             for angular integration of the Legendre coefficients.
#         print_level (bool) - print progress if true (useful to track long
#             property computations)
#     Result/Return:
#         bundle - reference to the input Bundle object. The property
#             key-0 is set
#     """
#
#     # Lebedev grid for S(\vec q)
#     leb = lebedev.Lebedev.build(nlebedev)
#     # Direct product grid for \vec q
#     qx = np.outer(q, leb.x)
#     qy = np.outer(q, leb.y)
#     qz = np.outer(q, leb.z)
#
#     for find, frame in enumerate(bundle.frames):
#         if print_level:
#             print 'Frame %5d of %5d' % (find, len(bundle.frames))
#         # Compute N(\vec q) = \sum_{A} f_A (\vec q) * \exp(-1.j * \vec q * \vec r)
#         N = np.zeros((len(q), len(leb.x)), dtype=complex)
#         for A, factor in enumerate(factors):
#             x = frame.xyz[A,0]
#             y = frame.xyz[A,1]
#             z = frame.xyz[A,2]
#             N += factor.evaluate_N(qx=qx,qy=qy,qz=qz,x=x,y=y,z=z)
#         # Compute I(\vec q) = N(\vec q)**2 for this frame
#         I = (np.abs(N)**2).real
#         # Compute Y00 (for common normalization)
#         Y00 = legendre.zonal2(leb.z, 0)[0]
#         # Integrate over S(2)
#         I0 = np.einsum('qw,w->q', I, leb.w * Y00)
#         # Assign the property to frame
#         frame.properties['%s-%d' % (key, 0)] = I0
#
#     return bundle



import os
import re

import numpy as np


# TODO: make energies frame property
def get_caspt2_energies(
    fn_in,
):

    ev_per_h = 27.211385

    # TODO: check whether computation finished
    all_lines = open(fn_in).readlines()
    energies = {}
    for n, line in enumerate(all_lines):
        mobj = re.match(r"^\s*\!RSPT2 STATE (\d)\.\d Energy\s*(\S+)", line)
        if mobj:
            state = int(mobj.group(1))
            energy = float(mobj.group(2))
            energies[state] = energy * ev_per_h

    return energies


def get_IP_molpro(
    out_dir,
    cation_out,
    neutral_out,
    traj,
    njobs,
):

    # TODO: verify frame corresponds to molpro run based on ID

    frames_per_job = len(traj.frames) / njobs
    nframes = len(traj.frames)
    for job in range(njobs + 1):
        os.chdir(out_dir)
        os.chdir("IP/%02djob" % job)
        ind1 = frames_per_job * job
        ind2 = frames_per_job * (job + 1)
        if job == njobs:
            ind2 = nframes
        for ind, frame in enumerate(traj.frames[ind1:ind2]):
            cEs = get_caspt2_energies(
                "%sIP/%02djob/%04d/%s" % (out_dir, job, ind + ind1, cation_out)
            )
            nEs = get_caspt2_energies(
                "%sIP/%02djob/%04d/%s" % (out_dir, job, ind + ind1, neutral_out)
            )
            nE = nEs[frame.I]
            IP = np.array([(key, nE - cE) for key, cE in list(cEs.items())])
            frame.properties["IP"] = IP
        os.chdir("../")
    os.chdir("../")

    return traj



import os

import numpy as np

import aimsprop as ai
import aimsprop.mananage_xyz as manage_xyz
from aimsprop import atom_data


def vis_aims(
    traj,
):

    # => Writing vmd scripts <= #
    os.system("mkdir xyzs")

    # => Write Bundle Files (for vmd to read) <= #
    filenames = []
    opacities = {}
    states = {}
    for index, label in enumerate(traj.labels):

        # all timesteps for trajectories with label x
        trajI = traj.subset_by_label(label)
        xyzs = np.array([frame.xyz for frame in trajI.frames])
        # Ns = np.array([frame.N for frame in trajI.frames])
        ts = np.array([frame.t for frame in trajI.frames])
        ws = np.array([frame.w for frame in trajI.frames])

        # creating and concatenating dummy frames
        # adding previous steps
        new_ts1 = np.arange(ts[1] - ts[0], ts[0], ts[1] - ts[0])
        new_ws1 = np.zeros_like(new_ts1)
        nts1 = len(new_ts1)
        natoms = np.shape(xyzs)[1]
        new_xyzs1 = np.zeros((nts1, natoms, 3))
        tot_xyzs = np.vstack((new_xyzs1, xyzs))
        tot_ws = np.hstack((new_ws1, ws))

        # output coordinates as xyz file for vmd
        symbols = [atom_data.atom_symbol_table[N] for N in frame.N]
        geoms = manage_xyz.xyzs_to_geom(symbols, tot_xyzs)
        manage_xyz.write_xyzs(
            "xyzs/x%02d-%02d.xyz" % (label[0], label[1]), geoms, scale=1.0
        )
        filenames.append("x%02d-%02d.xyz" % (label[0], label[1]))

        opacities[index] = tot_ws
        states[index] = frame.I - 1
        nframes = len(tot_ws)

    colors = [
        [1.0, 0.0, 0.0],
        [0.0, 0.0, 1.0],
        [0.0, 0.5, 0.5],
    ]
    ai.vis_aims.write_vmd(colors, filenames, states)
    options = {
        "opacities": opacities,
        "end_frame": nframes,
        "increment_frame": 1,
        "sleep": 0.5,
        "ntrajs": len(filenames),
    }
    ai.vis_aims.write_render(options)
    ai.vis_aims.run_vmd_render()
    ai.vis_aims.get_gif()


def test():

    # => Make a trajectory object <= #

    ICs = [3]
    trajs = [
        ai.parse_fms90("/home/monikaw/chem/o-np/6-aims/%04d/1-run" % x) for x in ICs
    ]

    # Merge the trajectories into one super-big Bundle with uniform weights
    traj = ai.Bundle.merge(trajs, [1.0 / len(trajs)] * len(trajs), labels=ICs)

    # Compute properties at ~1 fs intervals, removing nonsense due to adaptive timesteps
    ts = np.arange(np.min(traj.ts), np.max(traj.ts), 800.0)
    traj = traj.interpolate_linear(ts)
    print(f"nframes: {len(traj.ts)}")

    vis_aims(
        traj,
    )


if __name__ == "__main__":

    test()



# import os

import numpy as np

import aimsprop as ai


def test_major_bundle_operations(tmp_path, bundle):
    # 1. Plot Bond Distances (Spaghetti + Blur)
    # Compute the a bond distance property for all Frames in traj
    ai.compute_bond(bundle, "R01", 0, 1)
    ai.plot_scalar(
        tmp_path / "R.pdf",
        bundle,
        "R01",
        ylabel=r"$R_{CC} [\AA{}]$",
        time_units="fs",
        state_colors=["r", "b"],
        plot_average=True,
    )

    # Blur the bond distance (convolve)
    R = np.linspace(0.5, 3.0, 50)
    ai.blur_property(bundle, "R01", "Rblur", R, alpha=8.0)
    #    Plot the heat map of blurred bond distance
    ai.plot_vector(
        tmp_path / "Rblur.pdf",
        bundle,
        "Rblur",
        y=R,
        ylabel=r"$R [\AA{}]$",
        time_units="fs",
        nlevel=64,
    )

    # 2. Plot of Torsion Angle (Spaghetti + Blur)

    # Compute the a torsion angle property for all Frames in traj
    ai.compute_torsion(bundle, "T0123", 0, 1, 2, 3)
    ai.unwrap_property(bundle, "T0123", 360.0)
    ai.plot_scalar(
        tmp_path / "T.pdf",
        bundle,
        "T0123",
        ylabel=r"$\Theta [^{\circ{}}]$",
        time_units="fs",
        state_colors=["r", "b"],
        plot_average=True,
    )

    # Blur the torsion
    T = np.linspace(-180.0, +180.0, 100)
    ai.blur_property(bundle, "T0123", "Tblur", T, alpha=0.02)
    #    Plot the heat map of blurred torison
    ai.plot_vector(
        tmp_path / "Tblur.pdf",
        bundle,
        "Tblur",
        y=T,
        ylabel=r"$Theta [^{\circ{}}]$",
        time_units="fs",
        nlevel=64,
    )

    # 3. UED Cross Section

    # Compute the "simple" form of the UED cross section in R
    R = np.linspace(1.0, 6.0, 50)
    ai.compute_ued_simple(bundle, "UED", R=R, alpha=8.0)

    # Plot the heat map of the UED cross section detailed above
    ai.plot_vector(
        tmp_path / "UED.pdf",
        bundle,
        "UED",
        y=R,
        ylabel=r"$R [\AA{}]$",
        time_units="fs",
        diff=True,
    )



import pytest as pt

import aimsprop as ai


def test_compute_population(bundle):

    # 1. Compute the populations for states.
    # Test if the computed population at time 25 is correct.
    pop = ai.compute_population(bundle)
    assert pop[1][24] == pt.approx(0.441434205, rel=1e-7)



import numpy as np

import aimsprop as ai

# Parse a series of FMS90 trajectories that Hayley has run for Stilbene
trajs = [
    ai.parse_fms90("/home/hweir/stilbene/5-aims/s0_extension/aims_%04d/job1" % x)
    for x in [1, 2, 3]
]
# Merge the trajectories into one super-big Bundle with uniform weights
traj = ai.Bundle.merge(trajs, ws=[1.0 / len(trajs)] * len(trajs), labels=[1, 2, 3])
print((traj.labels))
# Compute properties at ~1 fs intervals, removing nonsense due to adaptive timesteps
ts = np.arange(0.0, max(traj.ts), 40.0)  # TODO: Cleaner edges
traj = traj.interpolate_nearest(ts)

# => Showoff Plot of Bond Distances (Spaghetti + Blur) <= #

# Compute the a bond distance property for all Frames in traj
# This particular bond is the one that leads to a three-ring complex
ai.compute_bond(traj, "R01", 7, 8)

# Blur the bond distance
R = np.linspace(1.0, 6.0, 50)
ai.blur_property(traj, "R01", "Rblur", R, alpha=8.0)

# Plot the heat map of blurred bond distance
# ai.plot_vector('R2.pdf', traj, 'Rblur', y=R, ylabel=r'$R [\AA{}]$', time_units='fs', twosided=False, cmap=plt.get_cmap('viridis'))
ai.plot_vector(
    "R.pdf", traj, "Rblur", y=R, ylabel=r"$R [\AA{}]$", time_units="fs", nlevel=64
)
ai.plot_scalar(
    "R.pdf",
    traj,
    "R01",
    ylabel=r"$R_{CC} [\AA{}]$",
    time_units="fs",
    state_colors=["r", "b"],
    clf=False,
    plot_average=False,
)

# => Showoff Plot of Torsion Angle (Spaghetti + Blur) <= #

# Compute the a torsion angle property for all Frames in traj
# This particular torsion is the one the leads to cis-trans isomerization
ai.compute_torsion(traj, "T4019", 4, 0, 1, 9)
ai.unwrap_property(traj, "T4019", 360.0)

# Blur the torsion
T = np.linspace(-180.0, +180.0, 100)
ai.blur_property(traj, "T4019", "Tblur", T, alpha=0.02)

# Plot the heat map of blurred torison
ai.plot_vector(
    "T.pdf",
    traj,
    "Tblur",
    y=T,
    ylabel=r"$Theta [^{\circ{}}]$",
    time_units="fs",
    nlevel=64,
)
ai.plot_scalar(
    "T.pdf",
    traj,
    "T4019",
    ylabel=r"$\Theta [^{\circ{}}]$",
    time_units="fs",
    state_colors=["r", "b"],
    clf=False,
    plot_average=False,
)

# => UED Cross Section <= #

# Compute the "simple" form of the UED cross section in R
R = np.linspace(1.0, 6.0, 50)
ai.compute_ued_simple(traj, "UED", R=R, alpha=8.0)

# Plot the heat map of the UED cross section detailed above
ai.plot_vector(
    "U.pdf", traj, "UED", y=R, ylabel=r"$R [\AA{}]$", time_units="fs", diff=True
)



import numpy as np

import aimsprop as ai

# => Parse/Align/Weight/Clean <= #

# Parse a series of FMS90 trajectories that Hayley has run for Stilbene
ICs = list(range(1, 16 + 1))
trajs = [
    ai.parse_fms90("/home/hweir/stilbene/5-aims/s0_extension/aims_%04d/job1" % x)
    for x in ICs
]
# Merge the trajectories into one super-big Bundle with uniform weights
traj = ai.Bundle.merge(trajs, ws=[1.0 / len(trajs)] * len(trajs), labels=ICs)
# Compute properties at ~1 fs intervals, removing nonsense due to adaptive timesteps
ts = np.arange(0.0, max(traj.ts), 40.0)
trajs = [traj2.interpolate_nearest(ts) for traj2 in trajs]
traj = traj.interpolate_nearest(ts)

# => Population Plot <= #

# If you want the detailed population curves (corresponding to traj.ts), e.g., for decay constant fitting:
# populations = ai.compute_population(traj)
# If you just want a plot of the populations
plt = ai.plot_population(
    "P.pdf",
    traj,
    trajs,
    time_units="fs",
    state_colors=["r", "b"],
)



import get_dyson
import get_molpro
import numpy as np

import aimsprop as ai

# => Parse/Align/Weight/Clean <= #

# Parse a series of FMS90 trajectories that Hayley has run for Stilbene
# Troubles: 11, 12 (not finished), 15
# trajs = [ai.parse_fms90('/home/hweir/stilbene/5-aims/s0_extension/aims_%04d/job1' % x) for x in range(1,16+1) if x not in [11, 12, 15]]
trajs = [
    ai.parse_fms90("/home/hweir/stilbene/5-aims/s0_extension/aims_%04d/job1" % x)
    for x in [1]
]
# TODO: Align trajectories to IC transition dipole moment (on z) and weight by oscillator strength at IC
# Merge the trajectories into one super-big Bundle with uniform weights
traj = ai.Bundle.merge(trajs, [1.0 / len(trajs)] * len(trajs))
# Compute properties at ~10 fs intervals, removing nonsense due to adaptive timesteps
ts = np.arange(0.0, max(traj.ts), 400.0)  # TODO: Cleaner edges
traj = traj.interpolate_nearest(ts)

# => Obtain Ionization Potential with Molpro (user defined process) <= #

# divide molpro calculations into reasonable number of jobs
njobs = 25

# molpro reference input files
neutral_input = "/home/monikaw/examples/pes_stilbene/ref/neutral.dat"
cation_input = "/home/monikaw/examples/pes_stilbene/ref/cation.dat"

# cation states to run, depending on frame state (1 = groundstate for both molpro and fms90)
# the active space for this system only permits 1 excited state
cation_states = {
    1: 1,
    2: 2,
}

# reference submission scripts
submit = "/home/monikaw/examples/pes_stilbene/ref/sbatch.sh"

# output working directory
out_dir = "/home/monikaw/examples/pes_stilbene/IP/"
opt_in = {
    "neutral_input": neutral_input,
    "cation_input": cation_input,
    "submit_input": submit,
    "out_dir": out_dir,
    "traj": traj,
    "cation_states": cation_states,
    "submit": False,
    "njobs": njobs,
}

# prep and / or run molpro jobs
# prep_molpro.prep_molpro(opt_in)

# assign ionization potentials as frame properties
# output working directory
out_dir = "/home/monikaw/examples/pes_stilbene/"
traj = get_molpro.get_IP_molpro(
    out_dir=out_dir,
    cation_out="cation.out",
    neutral_out="neutral.out",
    traj=traj,
    njobs=njobs,
)

# => Prepare Jobs for Calculation of Dyson Orbitals in Lightspeed <= #

out_dir_dyson = "/home/monikaw/examples/pes_stilbene/dyson/"
submit = "/home/monikaw/examples/pes_stilbene/ref/dyson/sbatch.sh"
opt_dyson = {
    "submit_input": submit,
    "out_dir": out_dir_dyson,
    "traj": traj,
    "submit": False,
    "njobs": njobs,
}
# prep and / or run dyson calculations
# prep_dyson.prep_dyson(opt_dyson)
traj = get_dyson.get_dyson(out_dir_dyson, traj, njobs)

# => Compute PES <= #

# kinetic energy of electron (in eV) (y axis of plot)
eKT = np.linspace(0.0, 15.0, 1000)
# experimental carrier frequency (eV)
carrier_frequency = 3.0
# gaussian blur, broadening parameter
alpha = 1.0
traj = ai.pes.compute_pes(
    traj=traj,
    carrier_frequency=carrier_frequency,
    alpha=alpha,
    eKT=eKT,
)

# => Plot Results <= #


import matplotlib.pyplot as plt

# change to plotting directory
filename = "/home/monikaw/examples/pes_stilbene/pes2D.pdf"
ai.plot.plot_vector(
    filename=filename,
    traj=traj,
    key="pes",
    y=eKT,
    ylabel="$\epsilon [eV]$",
    twosided=False,
    time_units="fs",
    nlevel=65,
    cmap=plt.cm.jet,
)



import numpy as np

from .bundle import Bundle


def compute_population(
    bundle: Bundle,
):
    """Compute the state populations for a bundle

    Params:
        bundle: the Bundle object to compute the property for
    Returns:
        populations: dictionary from I to time population of I. The indices of each vector in populations
        corresponds to the set of bundle.ts
    """

    ts = bundle.ts
    populations = {}
    for I in bundle.Is:
        bundle2 = bundle.subset_by_I(I)
        populations[I] = np.zeros((len(ts),))
        for tind, t in enumerate(ts):
            bundle3 = bundle2.subset_by_t(t)
            populations[I][tind] += sum([frame.w for frame in bundle3.frames])
    return populations



import pytest as pt

import aimsprop as ai


def test_compute_bond(bundle):

    # 1. Compute the bond distance between atoms 0 and 1 for the bundle.
    # Test if the first frame's bond length is correct.

    ai.compute_bond(bundle, "R01", 0, 1)
    assert bundle.frames[0].properties["R01"] == pt.approx(
        1.282298373, rel=1e-10
    )  # 0767554


def test_compute_angle(bundle):

    # 2. Compute the angle between atoms 0, 1, 2 for the bundle.
    #    Test if the first frame's angle is correct.

    ai.compute_angle(bundle, "A012", 0, 1, 2)
    assert bundle.frames[0].properties["A012"] == pt.approx(
        122.541809807, rel=1e-10
    )  # 4228


def test_compute_torsion(bundle):

    # 3. Compute the torsion between atoms 0, 1, 2, 3 for the bundle
    #    Test if the first frame's dihedral angle is correct.

    ai.compute_torsion(bundle, "T0123", 0, 1, 2, 3)
    assert bundle.frames[0].properties["T0123"] == pt.approx(
        160.465152603, rel=1e-10
    )  # 32674


def test_compute_oop(bundle):

    # 4. Compute the out of plane torsion between atoms 0, 1, 2, 3 for the bundle
    #    Test if the first frame's out of plane angle is correct.

    ai.compute_oop(bundle, "OOP0123", 0, 1, 2, 3)
    assert bundle.frames[0].properties["OOP0123"] == pt.approx(
        9.1198640147, rel=1e-11
    )  # 02102


def test_compute_transfer_coord(bundle):

    # 5. Compute the proton transfer coordinate between atoms 0, 1, 2, 3 for the bundle
    #   Test if the first frame's proton transfer coordinate matches the original

    ai.compute_transfer_coord(bundle, "PT012", 0, 1, 2)
    coordinate_0 = float(bundle.frames[0].properties["PT012"])
    assert coordinate_0 == pt.approx(-0.777663303, rel=1e-10)  # 0639367



"""
Written by Rob Parrish

This script reads TeraChem TDDFT and/or FOMO output files and automatically plots absorption spectra.

Each individual excited state is given a color, with the overall spectrum plotted in black on top.

The script includes options to plot with Lorentzian or Gaussian broadening. To change it, modify the ```plot_spectra``` function to change:

```python
Ostate += spectrum[state, 1] * lorentzian(E, spectrum[state, 0], delta)
```

to

```python
Ostate += spectrum[state, 1] * gaussian(E, spectrum[state, 0], delta)
```

It also includes options to plot in eV or nm, with `plot_spectra` or `plot_spectra_nm` respectively.

Modify the following code block with appropriate paths for your system, choice of fomo or tddft for `read_*_output`, and choice of eV of nm for `plot_*`.

```
filenames = glob.glob("/path/to/*/filename*.out")
spectra = [read_fomo_outfile(x) for x in filenames]
spectra = [x for x in spectra if x is not None]
plot_spectra("spectra.pdf", spectra, E=np.linspace(5.0, 9.0, 1000), delta=0.10)
```


"""
import glob
import re

import matplotlib
import matplotlib.pyplot as plt
import numpy as np

matplotlib.use("Agg")


# Python script to automatically extract TeraChem output (TD-DFT or CASCI) and plot
# Lorentzian or Gaussian broadened absorption spectrum.

# Can plot spectra in eV or nm, depending on how you modify the exection.


def read_tddft_outfile(filename):

    lines = open(filename).readlines()

    if len([x for x, y in enumerate(lines) if re.match(r"^\s+Job finished", y)]) == 0:
        return None

    lines = lines[
        (
            [
                x
                for x, y in enumerate(lines)
                if re.match(r"^\s+Final Excited State Results:", y)
            ][0]
            + 4
        ) :
    ]
    vals = []
    for line in lines:
        mobj = re.match(r"^\s*(\d+)\s+\S+\s+(\S+)\s+(\S+)", line)
        if mobj:
            vals.append([float(mobj.group(2)), float(mobj.group(3))])
            assert int(mobj.group(1)) == len(vals)
    return np.array(vals)


def read_fomo_outfile(filename):

    lines = open(filename).readlines()

    if len([x for x, y in enumerate(lines) if re.match(r"^\s+Job finished", y)]) == 0:
        return None

    Es = []
    re_ex = re.compile(r"^\s*(\d+)\s+singlet\s+\S+\s+\S+\s+(\S+)\s+\S+\s*$")
    for line in lines:
        mobj = re.match(re_ex, line)
        if mobj:
            Es.append(float(mobj.group(2)))
            assert int(mobj.group(1)) == len(Es) + 1

    Os = []
    re_osc = re.compile(r"^\s*1\s+->\s+(\d+)\s+\S+\s+\S+\s+\S+\s+\S+\s+(\S+)\s*$")
    for line in lines:
        mobj = re.match(re_osc, line)
        if mobj:
            Os.append(float(mobj.group(2)))
            assert int(mobj.group(1)) == len(Os) + 1
    assert len(Os) == len(Es)

    return np.array([[x, y] for x, y in zip(Es, Os)])


def lorentzian(
    x,
    x0,
    delta,
):

    return 0.5 * delta / np.pi * 1.0 / ((x - x0) ** 2 + (0.5 * delta) ** 2)


def gaussian(x, x0, delta):
    return np.exp(-np.power(x - x0, 2.0) / (2 * np.power(delta, 2.0)))


def plot_spectra(
    filename,
    spectra,
    E=np.linspace(3.5, 7.0, 1000),
    delta=0.05,
):

    Nspectra = len(spectra)
    Nstate = spectra[0].shape[0]

    Ostates = []
    for state in range(Nstate):
        Ostate = np.zeros_like(E)
        for spectrum in spectra:
            Ostate += spectrum[state, 1] * lorentzian(E, spectrum[state, 0], delta)
        Ostate /= Nspectra
        Ostates.append(Ostate)

    Ototal = np.zeros_like(E)
    for Ostate in Ostates:
        Ototal += Ostate

    cmap = matplotlib.cm.get_cmap("jet")
    colors = [cmap(float(x) / (Nstate - 1)) for x in reversed(list(range(Nstate)))]

    plt.clf()
    for Oind, Ostate in enumerate(Ostates):
        plt.plot(E, Ostate, color=colors[Oind])

    plt.plot(E, Ototal, "-k")
    plt.xlabel(r"$\Delta E$ [eV]")
    plt.ylabel("Cross Section [-]")
    plt.savefig(filename)


def plot_spectra_nm(
    filename,
    spectra,
    E=np.linspace(3.5, 7.0, 1000),
    delta=0.05,
):
    Nspectra = len(spectra)
    Nstate = spectra[0].shape[0]

    Ostates = []
    for state in range(Nstate):
        Ostate = np.zeros_like(E)
        for spectrum in spectra:
            Ostate += spectrum[state, 1] * lorentzian(E, spectrum[state, 0], delta)
        Ostate /= Nspectra
        Ostates.append(Ostate)

    Ototal = np.zeros_like(E)
    for Ostate in Ostates:
        Ototal += Ostate

    cmap = matplotlib.cm.get_cmap("jet")
    colors = [cmap(float(x) / (Nstate - 1)) for x in reversed(list(range(Nstate)))]

    nm = (1239.84193) / E

    plt.clf()
    for Oind, Ostate in enumerate(Ostates):
        plt.plot(nm, Ostate, color=colors[Oind])

    plt.plot(nm, Ototal, "-k")
    plt.xlabel(r"$\lambda$ [nm]")
    plt.ylabel("Cross Section [-]")
    plt.savefig(filename)


# --- Execution section --- #

if __name__ == "__main__":
    filenames = glob.glob("0*/output.dat")
    spectra = [read_fomo_outfile(x) for x in filenames]
    spectra = [x for x in spectra if x is not None]
    plot_spectra("spectra.pdf", spectra, E=np.linspace(5.0, 9.0, 1000), delta=0.10)



import matplotlib
import numpy as np

import aimsprop as ai

matplotlib.use("Agg")
import os

import matplotlib.pyplot as plt


def build_I2_traj(
    tmax=600.0,
    nframe=26,
    dz=0.5,
):

    """Provides a Bundle representing I2 aligned in z oscillating over one full cycle.

    Params:
        tmax (float) - maximum time
        nframe (int) - total number of frames
        dz (float) - amplitude of oscillation in Angstrom
    Returns:
        Bundle with I2 oscillating in z
    """

    # I2 in z
    N = [53] * 2
    xyz = np.array(
        [
            [0.000, 0.000, +1.33],
            [0.000, 0.000, -1.33],
        ]
    )

    ts = np.linspace(0.0, tmax, nframe)

    # Oscillate over tmax
    frames = []
    for t in ts:
        xyz2 = xyz + dz * np.sin(2.0 * np.pi * t / tmax) * np.array(
            [
                [0.0, 0.0, +1.0],
                [0.0, 0.0, -1.0],
            ]
        )
        frame = ai.Frame(
            label=0,
            t=t,
            w=1.0,
            I=1,
            N=N,
            xyz=xyz2,
        )
        frames.append(frame)

    traj = ai.Bundle(frames)

    return traj


def test_ued():

    """Demonstration of UED molecular movie computation.

    Steps:
    1 - get a Bundle
    2 - setup a grid in s and eta, for an e- beam with deBroglie wavelength L
    3 - compute diffraction moments I0(s, t) and I2(s, t)
    4 - compute diffraction pattern I(s, eta, t)
    5 - make detector movie of I(s, eta, t)
    6 - make plot of I0(s, t) [isotropic signal]
    7 - sine transform I0(s, t) -> R0(r, t)
    8 - make plot of R0(r, t) [molecular movie]
    9 - Gaussian blur R0(r, t) by 100 fs FWHM
    10 - make plot of R0blur(r, t) [molecular movie]

    As is standard practice, difference signals are used throughout. This
    removes background and signal due to unexcited background. Additionally,
    the modified diffraction signal s M(s) is used in place of I(s) throughout.
    Standard perpendicular pump-probe anisotropy (cos^2) is applied.

    The 3.7 MeV e- beam and s range used here is standard for the UED-I
    experiment at SLAC.
    """

    # => 1 - Bundle <= #

    # Get I2 oscillation trajectory
    traj = build_I2_traj(
        tmax=600.0,
        nframe=151,
        dz=0.5,
    )
    # Unique times in traj
    t = traj.ts

    # => 2 - Scattering Geometry <= #

    # 3.7 MeV electrons
    L = 0.003
    # Minimal s on detector (limited by e- beam hole)
    smin = 3.5
    # Maximal s on detector (limited by detector area and SNR)
    smax = 12.0
    # Sample s in [smin, smax]
    s = np.linspace(smin, smax, 120)
    # Detector azimuthal angles eta
    eta = np.linspace(0.0, 2.0 * np.pi, 61)
    # 2D detector grid
    ss, ee = np.meshgrid(s, eta, indexing="ij")
    # r to sample in sine transform
    r = np.linspace(0.0, 6.0, 120)

    # => 3 - Diffraction Moments <= #

    # Compute diffraction moments I-0 and I-2 (much faster than directly computing I(s, eta)
    traj = ai.iam.compute_diffraction_moments_fast(
        traj,
        "I",
        s=s,
        L=L,
        mode="ued",  # Set up fA(s) for UED
        nlebedev=4334,  # Very large solid angle grid
        nomega=1,  # Special case, on z, omega grid is not needed
        form="mod",  # Return M(s) instead of I(s)
        # print_level=True,
    )

    # => 4 - Diffraction Pattern <= #

    # Compute diffraction pattern I(s, eta) from I-0 and I-2
    ai.iam.compute_diffraction_from_moments(
        traj,
        "I",
        eta,
    )

    # => 5 - Detector Movie <= #

    # Lots of images for this -> place in movie dir
    dirname = "movie"
    if not os.path.exists(dirname):
        os.makedirs(dirname)

    # Get the difference detector patterns M[t,s,eta]
    I = traj.extract_property("I", diff=True)
    # sM version of difference detector patterns
    I *= np.einsum("i,jk->ijk", np.ones((I.shape[0],)), ss)

    # Make some nice levels for colormaps (uniform across t)
    Imax = np.max(np.abs(I))
    levels = np.linspace(-Imax, +Imax, 127)

    # Make a png image for each frame
    plt.clf()
    fig, ax = plt.subplots(subplot_kw=dict(projection="polar"))
    for tind, tval in enumerate(traj.ts):
        plt.cla()
        h = ax.contourf(
            ee,
            ss,
            I[tind, :, :],
            levels=levels,
            cmap=plt.get_cmap("bwr"),
            extend="both",
        )
        th = np.linspace(0.0, 2.0 * np.pi, 1000)
        ax.plot(th, smin * np.ones_like(th), "-k", linewidth=1.0)
        ax.plot(th, smax * np.ones_like(th), "-k", linewidth=1.0)
        ax.set_theta_offset(np.pi / 2.0)
        ax.set_rlim([0.0, smax])
        ax.set_rticks([])
        ax.set_thetagrids([])
        plt.axis("off")
        plt.savefig(
            "%s/I%03d.png" % (dirname, tind), bbox_inches="tight", transparent=True
        )

    # You can go into movie/ and stitch the png files together into a movie, e.g., with ffmpeg (check out movie.sh)

    # => 6 - Isotropic I0(s,t) Plot <= #

    # Get the isotropic part of the difference signal M0[t,s]
    I0 = traj.extract_property("I-0", diff=True)
    # Work with sM(s)
    I0 *= np.outer(np.ones_like(t), s)

    # Plot I0
    I0max = np.max(np.abs(I0))
    I0levels = np.linspace(-I0max, +I0max, 127)
    plt.clf()
    plt.contourf(t, s, I0.T, levels=I0levels, cmap=plt.get_cmap("bwr"))
    plt.xlabel("t [fs]")
    plt.ylabel("s [$\AA{}^{-1}$]")
    plt.colorbar()
    plt.savefig("I0.pdf", bbox_inches="tight")

    # => 7 - Sine Transform s -> r <= #

    # Compute the sine transform I0[t,s] -> R0[t,r]
    R0 = ai.compute_sine_transform(I0, s, r, a=7.0 ** (-2))

    # => 8 - Isotropic R0(r,t) Plot <= #

    R0max = np.max(np.abs(R0))
    R0levels = np.linspace(-R0max, +R0max, 127)
    plt.clf()
    plt.contourf(t, r, R0.T, levels=R0levels, cmap=plt.get_cmap("bwr"))
    plt.xlabel("t [fs]")
    plt.ylabel("r [$\AA{}$]")
    plt.colorbar()
    plt.savefig("R0.pdf", bbox_inches="tight")

    # => 9 - Gaussian Blurring in Time <= #

    # Blur in time to 100 fs FWHM
    R0blur = ai.compute_time_blur(
        R0,
        t,
        t,
        fwhm=100.0,
    )

    # => 10 - Isotropic R0blur(r,t) Plot <= #

    # Original levels
    R0max = np.max(np.abs(R0))
    R0levels = np.linspace(-R0max, +R0max, 127)
    plt.clf()
    plt.contourf(t, r, R0blur.T, levels=R0levels, cmap=plt.get_cmap("bwr"))
    plt.xlabel("t [fs]")
    plt.ylabel("r [$\AA{}$]")
    plt.colorbar()
    plt.savefig("R0blur.pdf", bbox_inches="tight")


if __name__ == "__main__":

    test_ued()



import numpy as np

import aimsprop as ai

# => Parse/Align/Weight/Clean <= #

# Parse a series of FMS90 trajectories that Hayley has run for Stilbene
# Troubles: 11, 12 (not finished), 15
# trajs = [ai.parse_fms90('/home/hweir/stilbene/5-aims/s0_extension/aims_%04d/job1' % x) for x in range(1,16+1) if x not in [11, 12, 15]]
trajs = [ai.parse_fms90("/home/parrish/chem/stil/6-aims/%04d" % x) for x in [1, 2]]
# TODO: Align trajectories to IC transition dipole moment (on z) and weight by oscillator strength at IC
# Merge the trajectories into one super-big Bundle with uniform weights
traj = ai.Bundle.merge(trajs, ws=[1.0 / len(trajs)] * len(trajs), labels=[1, 2])
# Compute properties at ~1 fs intervals, removing nonsense due to adaptive timesteps
ts = np.arange(0.0, max(traj.ts), 400.0)  # TODO: Cleaner edges
traj = traj.interpolate_nearest(ts)

# => Tag with X-Ray Scattering Signal <= #

# Grab form factors for all atoms in this trajectory
factors = ai.iam.AtomicFormFactor.build_factors(traj.frames[0], mode="xray")
# The q values to compute scattering cross section at (in A^-1)
q = np.linspace(0.5, 3.0, 100)
# Compute the diffraction pattern moments (l=0,2,4)
ai.iam.compute_diffraction_moments(
    traj=traj,
    key="xray",
    q=q,
    factors=factors,
    nlebedev=74,
    nlebedev2=74,
    nomega2=12,
    nlegendre=4,
    print_level=True,
)
# Plots of the result
ai.plot_vector(
    "I0.pdf",
    traj,
    "xray-0",
    y=q,
    ylabel=r"$q [\AA{}^{-1}]$",
    time_units="fs",
    diff=True,
)
ai.plot_vector(
    "I2.pdf",
    traj,
    "xray-2",
    y=q,
    ylabel=r"$q [\AA{}^{-1}]$",
    time_units="fs",
    diff=True,
)
ai.plot_vector(
    "I4.pdf",
    traj,
    "xray-4",
    y=q,
    ylabel=r"$q [\AA{}^{-1}]$",
    time_units="fs",
    diff=True,
)  # should be zero


# Compute the diffraction pattern moments (l=0 only)
ai.iam.compute_diffraction_moment0(
    traj=traj,
    key="xray0",
    q=q,
    factors=factors,
    nlebedev=74,
    print_level=True,
)
# Plots of the result
ai.plot_vector(
    "I00.pdf",
    traj,
    "xray0-0",
    y=q,
    ylabel=r"$q [\AA{}^{-1}]$",
    time_units="fs",
    diff=True,
)



from pathlib import Path

from scripts.convert_to_geomdat import convert_geometries


def test_convert_geometries(tmp_path):
    # Get input and answer files, sorted by name
    test_data_dir = Path(__file__).parent / "test_data" / "scripts"
    prmtop_file = test_data_dir / "parm.prmtop"
    rst7_files = sorted(list(test_data_dir.glob("**/coors_*.rst7")))
    correct_geom_files = sorted(list(test_data_dir.glob("**/Geometry_*.dat")))

    # Exercise the convert_geometries function
    for rst7 in rst7_files:
        convert_geometries(prmtop_file, rst7, tmp_path)

    # Collect the created files
    outputs = sorted(list(tmp_path.glob("**/*.dat")))

    # Check that the created files match the correct Geometry.dat files
    for i, output in enumerate(outputs):
        with open(output) as f:
            created = f.read()
        with open(correct_geom_files[i]) as f:
            correct = f.read()

        # Assert that each created files matches the corresponding answer key file
        assert created == correct



import numpy as np

from .bundle import Bundle


def blur_property(
    bundle: Bundle,
    key: str,
    key2: str,
    R: np.ndarray,
    alpha: float,
) -> Bundle:

    """Blur a property via Gaussian convolution (blurring applied in the "spatial" coordinate).

    Pararms:
        bundle: the Bundle object to compute the property for (modified in
            place)
        key: the name of the original property
        key2: the name of the blurred property
        R2: the grid of property values to blur to (usually R or
            theta or Q).
        alpha: the Gaussian blurring exponent
    Result/Return:
        bundle (Bundle): reference to the input Bundle object. The
            property key2 is set to the np.ndarray value of the blurred
            property.

    """

    for frame in bundle.frames:
        V = frame.properties[key]
        V2 = np.zeros_like(R)
        if np.array(V).ndim == 0:
            # Yak shave, I hates it!!
            V2 += np.sqrt(alpha / np.pi) * np.exp(-alpha * (R - V) ** 2)
        else:
            for RAB in V:
                V2 += np.sqrt(alpha / np.pi) * np.exp(-alpha * (R - RAB) ** 2)
        frame.properties[key2] = V2
    return bundle


def compute_time_blur(
    I: np.ndarray,
    t1: np.ndarray,
    t2: np.ndarray,
    fwhm: float,
) -> np.ndarray:

    """Compute Gaussian blurring in time for an arbitrary property.

    Uses the trapezoid rule over provided t1 (t1/t2 need not be uniformly spaced).

    Params:
        I: signal to Gaussian blur in the 0-th dim, shape (nt1, ...).
        t1: t1 values (can be irregular), shape (nt1,).
        t2: t2 values (can be irregular), shape (nt2,).
        fwhm: the full width at half maximum (FWHM) of the blurring
            kernel.
    Returns:
        I2 - Gaussian blurred signal, shape (nt2, ...).
    """

    # Irregular trapezoid weights in t1
    dt = np.diff(t1)
    w = np.zeros_like(t1)
    w[:-1] += 0.5 * dt
    w[1:] += 0.5 * dt

    # Gaussian exponent corresponding to fwhm
    a = 4.0 * np.log(2.0) / (fwhm ** 2)

    # Blurring kernel
    tt1, tt2 = np.meshgrid(t1, t2, indexing="ij")
    K = (a / np.pi) ** (0.5) * np.exp(-a * (tt1 - tt2) ** 2)

    if I.ndim == 2:
        V = I * np.outer(w, np.ones((I.shape[1],)))
        I2 = np.dot(K, V)
    else:
        raise ValueError("ndim case %d not coded" % I.ndim)

    return I2



import re

"""

Contains dictionary of units conversions to and from atomic units

"""


units = {
    "au_per_ang": 1.8897261328856432e00,  # au of length per Angstrom, PSI4
    "au_per_amu": 1.8228884840986366e03,  # au of mass per atomic mass unit, PSI4
    "au_per_debeye": 3.9343030716327643e-01,  # au of dipole per Debeye, PSI4
    "au_per_ev": 3.6749330610942925e-02,  # au of energy per eV, PSI4
    "au_per_kcal": 1.5936013717720609e-03,  # au of energy per kcal mol^-1, PSI4
    "au_per_cminv": 4.5563359040180500e-06,  # au of energy per cm^{-1} ("wavenumber"), PSI4
    "au_per_K": 1.0
    / 3.1577464e5,  # temperature (TODO: This is probably OK, maybe a bit low precision)
    "au_per_fs": 1.0 / 2.418884326505e-2,  # time
}
for k in list(units.keys()):
    v = units[k]
    mobj = re.match("(\S+)_per_(\S+)", k)
    units["%s_per_%s" % (mobj.group(2), mobj.group(1))] = 1.0 / v



import pytest as pt

import aimsprop as ai


def test_rotate_frames(bundle):

    # 1. Rotate the coordinates according to conversion matrix R.
    #    Test if the first rotated coordinate is correct.

    R = [[0, 1, 0], [1, 0, 1], [0, 1, 1]]
    ai.rotate_frames(bundle, R)
    assert bundle.frames[0].xyz[0][0] == pt.approx(0.035608866, rel=1e-9)


def test_rotate_frames_to_z(bundle):

    # 1. Rotate the coordinates according to the z vector for TDP.
    #    Test if the first rotated coordinate is correct.

    z = [0.2, 0.1, 0.1]
    ai.rotate_frames_to_z(bundle, z)
    assert bundle.frames[0].xyz[0][0] == pt.approx(-0.2099543620557265, rel=1e-9)



import re
from pathlib import Path
from typing import Any, Dict, List

import numpy as np

from . import atom_data, bundle

_N_table = {val: key for key, val in list(atom_data.atom_symbol_table.items())}


def _get_file_index(filename: Path) -> int:
    """
    extracts the number X from filename fanculo.X.ext
    """
    return int(filename.name.split(".")[1])


def _parse_positions(filepath: Path, cutoff_time: float = None) -> Dict[Any, Any]:
    """Parse information in positions.*.xyz.

    Arguments:
        filepath: the path to the FMS run directory
        cutoff_time: cutoff time to stop reading bundle info after
            (None reads all times).
    Returns:
        (N2s, xyz2s): Tuple of atomic indices and list of xyzs for each timme and state (dict of dict)
    """

    posfiles = (
        (_get_file_index(path), path)
        for path in filepath.iterdir()
        if path.match("positions.*.xyz")
    )

    # Read the Amp and positions files
    N2s = {}  # dict of Ns
    xyz2s = {}  # dict of xyzs
    for index, posfile in posfiles:
        Ns = {}
        xyzs = {}
        with open(posfile) as f:
            lines = f.readlines()
        natom = int(lines[0])
        if len(lines) % (natom + 2) != 0:
            raise RuntimeError("Position file does not have correct number of lines.")
        nframe = len(lines) // (natom + 2)
        for A in range(nframe):
            lines2 = lines[A * (natom + 2) : (A + 1) * (natom + 2)]
            mobj = re.match(r"^\s*Time:\s+(\S+),\s+Trajectory:\d+\s*$", lines2[1])
            t = float(mobj.group(1))
            if cutoff_time and t > cutoff_time:
                continue
            N = []
            xyz = np.zeros((natom, 3))
            for A, line in enumerate(lines2[2:]):
                mobj = re.match(r"^\s*(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s*$", line)
                N.append(_N_table[mobj.group(1).upper()])
                xyz[A, 0] = float(mobj.group(2))
                xyz[A, 1] = float(mobj.group(3))
                xyz[A, 2] = float(mobj.group(4))
            Ns[t] = N
            xyzs[t] = xyz
        N2s[index] = Ns
        xyz2s[index] = xyzs

        # N2s have the list of atoms repeated - why???

    return N2s, xyz2s


def _parse_Cs(filepath: Path, cutoff_time: float = None) -> Dict[Any, Any]:
    """Parse information in Amps.*.

    Arguments:
        filepath: the path to the FMS run directory
        cutoff_time: cutoff time to stop reading bundle info after
            (None reads all times).
    Returns:
        C2s: Amps (dict of dicts)

    """

    Cfiles = (
        (_get_file_index(path), path)
        for path in filepath.iterdir()
        if path.match("Amp.*")
    )
    C2s = {}
    for index, Cfile in Cfiles:

        # AV 2021, I do not want to further modify this parser, but of course it is stupid
        # to read the file in lines and THEN loop over them.
        with open(Cfile) as f:
            lines = f.readlines()[1:]

        Cs = {}
        for line in lines:
            # (t) Norm (Real) (Imag)
            mobj = re.match(r"^\s*(\S+)\s+\S+\s+(\S+)\s+(\S+)\s*$", line)
            if mobj:
                if cutoff_time and float(mobj.group(1)) > cutoff_time:
                    continue
                Cs[float(mobj.group(1))] = complex(
                    float(mobj.group(2)), float(mobj.group(3))
                )
            else:
                raise RuntimeError(f"Invalid Amp line: {line}")

        C2s[index] = Cs

    return C2s


def _parse_Ss(filepath: Path, cutoff_time: float = None) -> Dict[Any, Any]:
    """Parse information in S.dat.

    Arguments:
        filepath: the path to the FMS run directory
        cutoff_time: cutoff time to stop reading bundle info after
            (None reads all times).
    Returns:
        S2s: Overlap matrix (dict of dicts)
    """

    Sfile = filepath / "S.dat"

    # Read the overlap matrix
    Ss = {}
    with open(Sfile) as f:
        lines = f.readlines()
    tinds = []
    ts = []
    for lind, line in enumerate(lines):
        mobj = re.match(r"^\s*(\S+)\s*$", line)
        if mobj:
            tinds.append(lind)
            ts.append(float(mobj.group(1)))
    tinds.append(len(lines))
    for A, t in enumerate(ts):
        if cutoff_time and t > cutoff_time:
            continue
        lines2 = lines[tinds[A] + 1 : tinds[A + 1]]
        # This thing wraps at a certain point.
        # Just read in a 1D array, and then reshape to 2D
        Spart = []
        for line in lines2:
            toks = line.split()
            for B in range(len(toks)):
                if B % 2 != 0:
                    continue
                Sval = complex(float(toks[B]), float(toks[B + 1]))
                Spart.append(Sval)
        Smat = np.array(Spart)
        # Reshape to square array
        n = int(np.sqrt(len(Smat)))
        if n ** 2 != len(Smat):
            raise RuntimeError("S matrix is not square")
        Smat = np.reshape(Smat, (n, n))
        Ss[t] = Smat

    return Ss


def _parse_spawnlog(filepath: Path, initial_I: int = None) -> Dict[Any, Any]:
    """Parse information in Spawn.log.

    Arguments:
        filepath: the path to the FMS run directory
    Returns:
        states: electronic states
    """

    Spawnfile = filepath / "Spawn.log"

    # Read the Spawn.log to figure out electronic states (not read
    states = {}
    # AV: why this if? Why don't we ALWAYS use the Amp files to get the states variable?
    if not Spawnfile.is_file():
        if initial_I is None:
            raise RuntimeError("No Spawn.log and no initial_I")
        Cfiles_indexes = (
            _get_file_index(path) for path in filepath.iterdir() if path.match("Amp.*")
        )
        states = {k: initial_I for k in Cfiles_indexes}
    else:
        with open(Spawnfile) as f:
            lines = f.readlines()[1:]
        # CB: if we have the Spawn.log, read its content
        for lind, line in enumerate(lines):
            # match groups are TBF ID, state, parent TBF ID, parent TBF state
            mobj = re.match(r"^\s*\S+\s+\S+\s+\S+\s+(\d+)\s+(\d+)\s+(\d+)\s+(\d+)", line)
            states[int(mobj.group(1))] = int(mobj.group(2))
            # Redundant, but captures IC TBFs
            states[int(mobj.group(3))] = int(mobj.group(4))

    return states


def _swap_dic_axis(x2s: Dict[Any, Any]) -> Dict[Any, Any]:
    """Swap axis of dict to make it faster"""
    x3s = {}
    for index, x2 in list(x2s.items()):
        for t, x in list(x2.items()):
            x3s.setdefault(t, {})[index] = x

    return x3s


def _create_frames_mulliken(
    Ss: Dict[Any, Any],
    C3s: Dict[Any, Any],
    xyz3s: Dict[Any, Any],
    N3s: Dict[Any, Any],
    states: Dict[Any, Any],
    widths: List[float],
) -> List[bundle.Frame]:
    """Build list of Frames from parsed data using mulliken scheme

    Arguments:
        Ss (dic): Overlap
        C3s (dic): Amps
        xyz3s (dic): positions
        N3s (dic): atomic numbers
        states: electronic states
        widths: the gaussian widths of the atoms in this frame,

    Returns:
        Frames: list of frame objects
    """
    frames = []
    for t, S in list(Ss.items()):
        if t not in C3s:
            # Sometimes timestamps do not match because Amp.* only holds 2 decimal digits
            # E.g., 1000.875 (in S) vs. 1000.88 (in Amp)
            print(
                f"Warning: Time {t} not in amplitudes (OK if very small adaptive timestep)"
            )
            continue

        Cs = C3s[t]
        xyzs = xyz3s[t]
        Ns = N3s[t]

        if len(Cs) != len(S):
            raise RuntimeError("C and S are different sizes")

        Isort = list(sorted(Cs.keys()))

        for I2, I in enumerate(Isort):
            q = 0.0
            for J2, J in enumerate(Isort):
                if states[I] != states[J]:
                    continue  # Electronic orthogonality
                q += np.real(
                    0.5 * np.conj(Cs[I]) * S[I2, J2] * Cs[J]
                    + 0.5 * np.conj(Cs[J]) * S[J2, I2] * Cs[I]
                )
            frame = bundle.Frame(
                label=I,
                t=t,
                w=q,
                I=states[I],
                N=Ns[I],
                xyz=xyzs[I],
                widths=widths,
            )
            frames.append(frame)

    return frames


def _create_frames_saddle(
    Ss: Dict[Any, Any],
    C3s: Dict[Any, Any],
    xyz3s: Dict[Any, Any],
    N3s: Dict[Any, Any],
    states: Dict[Any, Any],
    cutoff_saddle: float,
    widths: List[float],
) -> List[bundle.Frame]:
    """Build list of Frames from parsed data using saddle scheme

    Arguments:
        Ss (dic): Overlap
        C3s (dic): Amps
        xyz3s (dic): positions
        N3s (dic): atomic numbers
        states: electronic states
        cutoff_saddle (float): cutoff for centroid TBF pair in the saddle point approach.
    Returns:
        Frames: list of frame objects
    """

    frames = []
    for t, S in list(Ss.items()):
        if t not in C3s:
            # Sometimes timestamps do not match because Amp.* only holds 2 decimal digits
            # E.g., 1000.875 (in S) vs. 1000.88 (in Amp)
            print(
                f"Warning: Time {t} not in amplitudes (OK if very small adaptive timestep)"
            )
            continue

        Cs = C3s[t]
        xyzs = xyz3s[t]
        Ns = N3s[t]

        if len(Cs) != len(S):
            raise RuntimeError("C and S are different sizes")

        Isort = list(sorted(Cs.keys()))

        for I2, I in enumerate(Isort):
            for J2, J in enumerate(Isort):
                if states[I] != states[J]:
                    continue  # Electronic orthogonality
                if J > I:
                    continue
                q = np.real(
                    0.5 * np.conj(Cs[I]) * S[I2, J2] * Cs[J]
                    + 0.5 * np.conj(Cs[J]) * S[J2, I2] * Cs[I]
                )
                if J < I:
                    q *= 2.0  # Upper/lower triangle
                if abs(q) < cutoff_saddle:
                    continue  # Vanishing weight
                frame = bundle.Frame(
                    label=(I, J),
                    t=t,
                    w=q,
                    I=states[I],
                    N=Ns[I],
                    xyz=0.5 * (xyzs[I] + xyzs[J]),  # centroid
                    widths=widths,
                )
                frames.append(frame)

    return frames


def _parse_widths(filepath) -> List[float]:
    """
    parse the FMS.out tio get widths
    """
    fms_out = filepath / "FMS.out"
    widths = []

    if fms_out.exists():
        print(f"reading {fms_out}")
        with open(fms_out, "r") as f:
            for line in f:
                if re.search("Width:", line):
                    a = float(line.split()[1])
                    widths.append(a)
    return widths


def parse_fms90(
    filepath: Path,
    scheme: str,
    cutoff_time: float = None,
    cutoff_saddle: float = 1.0e-4,
    initial_I: int = None,
) -> bundle.Bundle:
    """Parse an FMS90 results directory into a Bundle.

    This uses information in positions.*.xyz, Amps.*, S.dat, FMS.out and Spawn.log to
    populate a density matrix by mulliken or saddle point rules.

    Arguments:
        filepath: the path to the FMS run directory
        scheme: 'mulliken' or 'saddle' to indicate the approximation
            used for property evaluation.
        cutoff_time: cutoff time to stop reading bundle info after
            (None reads all times).
        cutoff_saddle: cutoff for centroid TBF pair in the saddle
            point approach.
        initial_I: initial electronic state, used only if there is
            no Spawn.log (e.g., if no spawning has happened yet) to place
            electronic label.
    Returns:
        Bundle: The Bundle object.
    """

    # Read in FMS output files: positions*, Amp.*, S.dat and Spawn.log
    N2s, xyz2s = _parse_positions(filepath, cutoff_time)

    C2s = _parse_Cs(filepath, cutoff_time)

    Ss = _parse_Ss(filepath, cutoff_time)

    states = _parse_spawnlog(filepath, initial_I)

    widths = _parse_widths(filepath)
    if widths == []:  # this code is beautiful.
        widths = atom_data.from_Ns_to_widths(
            N2s[list(N2s.keys())[0]][list(N2s[list(N2s.keys())[0]].keys())[0]]
        )

    if len(C2s) != len(N2s):
        raise RuntimeError("xyz and C files not same number of TBF")

    # Swap to put time on slow axis (C3s[t][I] instead of C2s[I][t])
    C3s, xyz3s, N3s = [_swap_dic_axis(x) for x in [C2s, xyz2s, N2s]]

    if scheme == "mulliken":
        frames = _create_frames_mulliken(Ss, C3s, xyz3s, N3s, states, widths)
    elif scheme == "saddle":
        frames = _create_frames_saddle(
            Ss, C3s, xyz3s, N3s, states, cutoff_saddle, widths
        )
    else:
        raise RuntimeError(f"Invalid scheme: {scheme}")

    processed_bundle = bundle.Bundle(frames)

    return processed_bundle



#!/usr/bin/env python
# setup.py generated by flit for tools that don't yet use PEP 517

from distutils.core import setup

packages = \
['aimsprop', 'aimsprop.iam', 'aimsprop.iam.ext', 'aimsprop.vis_aims']

package_data = \
{'': ['*'], 'aimsprop.vis_aims': ['ref/*']}

install_requires = \
['numpy', 'matplotlib']

extras_require = \
{'dev': ['black>=20.8b1',
         'flake8 >=3.8.4',
         'isort >= 5.7.0',
         'pre-commit >= 2.11.1',
         'autoflake >= 1.4'],
 'doc': ['mkdocs == 1.1.2',
         'mkdocs-material == 7.0.3',
         'mkdocstrings == 0.15.0'],
 'test': ['pytest >=6.2.2',
          'pytest-cov >=2.10.1,<3.0.0',
          'coverage >=5.3.1,<6.0']}

setup(name='aimsprop',
      version='0.2.1',
      description='A repository for the representation and manipulation of AIMS-type trajectories.',
      author='Rob Parrish',
      author_email='mtzgroupadmin@gmail.com',
      url='https://mtzgroup.github.io/aimsprop/',
      packages=packages,
      package_data=package_data,
      install_requires=install_requires,
      extras_require=extras_require,
      python_requires='>=3.6',
     )



import os

import aimsprop.atom_data as atom_data
import aimsprop.manage_xyz as manage_xyz

# TODO: make more invariant to filename (sbatch.sh)

# run molpro calculations to obtain ionization potential information
def prep_dyson(
    opt_in,
):

    """Runs lightspeed calculations from template files and traj object

    Params:
        opt_in - [dict], options to override defaults

    """

    options = {
        # all paths must be absolute!
        "submit_input": None,  # sbatch file to edit
        "submit_input2": "sbatch.sh",  # modified sbatch file to execute
        "out_dir": None,  # dir for output
        "submit": False,  # whether to submit or just prep the job
        "traj": None,  # traj object to run (all frames)
        "njobs": 25,  # number of jobs
    }

    for key, val in list(opt_in.items()):
        options[key] = val

    # TODO: throw error messages for invalid combination of options
    # override options
    submit_input = options["submit_input"]
    submit_input2 = options["submit_input2"]

    out_dir = options["out_dir"]
    submit = options["submit"]
    traj = options["traj"]
    njobs = options["njobs"]

    # loop over jobs
    frames_per_job = len(traj.frames) / njobs
    nframes = len(traj.frames)
    for job in range(njobs + 1):

        # setup directories
        os.chdir(out_dir)
        os.system("mkdir %02djob" % job)
        os.system("cp %s %02djob" % (submit_input, job))
        os.system("cp ../ref/dyson_fomo.py %02djob" % job)
        os.chdir("%02djob" % job)

        # divide frames per job (nframes % njobs goes into last job)
        ind1 = frames_per_job * job
        ind2 = frames_per_job * (job + 1)
        if job == njobs:
            ind2 = nframes

        states = [frame.I - 1 for frame in traj.frames[ind1:ind2]]
        nstates = "[ "
        for state in states:
            nstates += str(state) + ", "
        nstates += "]"
        # update submission script
        update_sbatch("%d" % job, submit_input2, ind1=ind1, ind2=ind2)
        update_dyson(nstates=nstates)

        # loop over frames
        for ind, frame in enumerate(traj.frames[ind1:ind2]):
            # set up directory
            os.system("mkdir %04d" % (ind + ind1))
            os.chdir("%04d" % (ind + ind1))

            # modify neutral molpro params
            symbols = [atom_data.atom_symbol_table[N] for N in frame.N]
            write_xyz(frame, symbols, "geom.xyz")

            # create file to identify frame with job
            os.system(
                "echo 'labels: %4d %4d\nt: %12.6f' > %s"
                % (frame.label[0], frame.label[1], frame.t, "%04d_ID.txt" % ind)
            )

            # return to working directory
            os.chdir("../")

        # if requested, submit the job
        if submit == True:
            os.system("sbatch %s" % submit_input2)
        os.chdir("../")


def update_sbatch(
    new_txt,
    sbatch_in,
    ind1,
    ind2,
):

    # jobname
    os.system('sed -i "s/ID/%s/g" %s' % (new_txt, sbatch_in))

    # frame indices
    os.system('sed -i "s/ind1/%s/g" %s' % (ind1, sbatch_in))
    os.system('sed -i "s/ind2/%s/g" %s' % (ind2, sbatch_in))


def update_dyson(
    nstates,
    filename="dyson_fomo.py",
):

    os.system('sed -i "s/neutral_state_values/%s/g" %s' % (nstates, filename))


def write_xyz(
    frame,
    symbols,
    filename,
):

    geom = []
    for ind, atom in enumerate(frame.xyz):
        geom.append((symbols[ind], atom[0], atom[1], atom[2]))
    manage_xyz.write_xyz(filename, geom, scale=1.0)



import numpy as np


def compute_sine_transform(
    I: np.ndarray,
    s: np.ndarray,
    r: np.ndarray,
    a: float = 0.0,
) -> np.ndarray:
    """Compute the sine transform I(t,r) = \int_0^smax ds I(t,s) * sin(r*s) * exp(-a * s^2)

    Uses the trapezoid rule over provided s (s need not be uniformly spaced).
    The factor of exp(-a * s^2) is a common experimental cutoff factor applied
    to prevent ringing from noise at large s. This tends to smooth the r-domain
    signal, but can limit the resolution.

    Params:
        I: signal to sine transform, e.g.,
            sM(t,s). Time should be on dim 0, s should be on dim 1,
            shape (nt, ns).
        s: s values (can be irregular), shape (ns,)
        r: r values (can be irregular), shape (nr,)
        a: Gaussian exponent in large-s cutoff Gaussian.
    Returns:
        I2: sine transformed signal, shape (nt, nr)
    """

    # Irregular trapezoid weights
    ds = np.diff(s)
    w = np.zeros_like(s)
    w[:-1] += 0.5 * ds
    w[1:] += 0.5 * ds

    V = I * np.outer(np.ones((I.shape[0],)), w * np.exp(-a * s ** 2))
    rr, ss = np.meshgrid(r, s, indexing="ij")
    K = np.sin(rr * ss)
    I2 = np.dot(V, K.T)
    return I2



import os
import re

import numpy as np

from . import atom_data, bundle

# TODO: Maybe should be in atom data
_N_table = {val: key for key, val in list(atom_data.atom_symbol_table.items())}


def parse_xyz(
    filename: str,
    label=1,
    w=1.0,
    I=1,
    t0=0.0,
    dt=20.0,
    ts=None,
    N_table=None,
) -> bundle.Bundle:

    """Parse an XYZ adiabatic bundle file directly into a Bundle.

        filename (str): the absolute or relative path to the xyz file.
        label (hashable): the label of this bundle
        w (float): the weight of this bundle
        I (int): electronic state label
        t0 (float): the initial time in au
        dt (float): the timestep in au
        ts (list of float): an explicit list of times in au, overrides t0 and dt
        N_table (dict of str : int): an optional dictionary mapping atomic
            symbol to atomic number, used for non-standard atom names.

    Returns:
        bundle (Bundle): the Bundle object.
    """

    lines = open(filename).readlines()

    natom = int(lines[0])  # This should always work
    if len(lines) % (natom + 2):
        raise ValueError("Invalid number of lines in xyz file")
    nframe = len(lines) / (natom + 2)

    xyzs = []
    Zs = []
    for frame in range(nframe):
        lines2 = lines[frame * (natom + 2) + 2 : (frame + 1) * (natom + 2)]
        Z = []
        xyz = []
        for line in lines2:
            mobj = re.match(r"^\s*(\S+)\s+(\S+)\s+(\S+)\s+(\S+)\s*$", line)
            Z.append(mobj.group(1))
            xyz.append([float(mobj.group(x)) for x in (2, 3, 4)])
        xyz = np.array(xyz)
        xyzs.append(xyz)
        Zs.append(Z)

    # User symbol table or default?
    N_table2 = N_table if N_table else _N_table

    frames2 = []
    for ind, xyz in enumerate(xyzs):
        Z = Zs[ind]
        Ns = [N_table2[key] for key in Z]
        widths = atom_data.from_Ns_to_widths(Ns)
        frame2 = bundle.Frame(
            label=label,
            t=dt * ind + t0 if ts is None else ts[ind],
            w=w,
            I=I,
            N=Ns,
            xyz=xyz,
            widths=widths,
        )
        frames2.append(frame2)

    parsed_bundle = bundle.Bundle(frames2)

    return parsed_bundle


def write_xyzs(
    bundle: bundle.Bundle,
    dirname: str,
    atom_format_str: str = "%-3s %24.16E %24.16E %24.16E\n",
):

    """Write a directory of xyz files to represent a Bundle, with
        one xyz file containing all frames for each label

    Params:
        bundle: Bundle to write xyz file representation of
        dirname: the directory to place the xyz files in (created if does not exist)
        atom_format_str: the format string for each atom line in the xyz
            file (useful to change precision).
    Result:
        xyz files are written for each label in bundle. Each xyz
        file contains all frames for the label, in time-order
    """

    # Make sure directoy exists
    if not os.path.exists(dirname):
        os.makedirs(dirname)

    # Write xyz files
    for label in bundle.labels:
        bundle2 = bundle.subset_by_label(label)
        xyzfilename = str(label)
        # Munging with filename label
        xyzfilename = xyzfilename.replace(" ", "")
        xyzfilename = xyzfilename.replace("(", "")
        xyzfilename = xyzfilename.replace(")", "")
        xyzfilename = xyzfilename.replace(",", "-")
        fh = open("%s/%s.xyz" % (dirname, xyzfilename), "w")
        for frame in bundle2.frames:
            fh.write("%d\n" % frame.xyz.shape[0])
            fh.write(
                "t = %24.16E, w = %24.16E, I = %d\n"
                % (
                    frame.t,
                    frame.w,
                    frame.I,
                )
            )
            for A in range(frame.xyz.shape[0]):
                fh.write(
                    atom_format_str
                    % (
                        atom_data.atom_symbol_table[frame.N[A]],
                        frame.xyz[A, 0],
                        frame.xyz[A, 1],
                        frame.xyz[A, 2],
                    )
                )
